@@ -16,28 +16,30 @@
 #include <unistd.h>
 #include <string.h>
 #include <limits.h>
-#include "ViennaRNA/fold.h"
-#include "ViennaRNA/part_func.h"
 #include "ViennaRNA/fold_vars.h"
+#include "ViennaRNA/eval/structures.h"
 #include "ViennaRNA/plotting/alignments.h"
 #include "ViennaRNA/plotting/structures.h"
 #include "ViennaRNA/plotting/utils.h"
 #include "ViennaRNA/utils/basic.h"
 #include "ViennaRNA/utils/strings.h"
+#include "ViennaRNA/utils/log.h"
 #include "ViennaRNA/alifold.h"
-#include "ViennaRNA/Lfold.h"
-#include "ViennaRNA/utils/alignments.h"
+#include "ViennaRNA/mfe/local.h"
+#include "ViennaRNA/sequences/alignments.h"
 #include "ViennaRNA/params/io.h"
-#include "ViennaRNA/alphabet.h"
+#include "ViennaRNA/sequences/alphabet.h"
 #include "ViennaRNA/io/file_formats_msa.h"
 #include "ViennaRNA/io/utils.h"
 #include "ViennaRNA/constraints/basic.h"
-#include "ViennaRNA/constraints/SHAPE.h"
-#include "RNALalifold_cmdl.h"
-#include "gengetopt_helper.h"
-#include "input_id_helpers.h"
+#include "ViennaRNA/probing/SHAPE.h"
 
-#include "ViennaRNA/color_output.inc"
+#include "RNALalifold_cmdl.h"
+#include "gengetopt_helpers.h"
+#include "input_id_helpers.h"
+#include "probing_data_helpers.h"
+
+#include "ViennaRNA/intern/color_output.h"
 
 #define DEFAULT_SPAN  70;
 
@@ -50,6 +52,8 @@ typedef struct {
   vrna_md_t *md;
   int       ss_eps;
   int       msa_eps;
+  double    color_threshold;
+  double    color_min_sat;
   int       msa_stk;
   int       csv;
   int       mis;
@@ -62,11 +66,11 @@ typedef struct {
 
 
 PRIVATE void
-print_hit_cb(int        start,
-             int        end,
-             const char *structure,
-             float      en,
-             void       *data);
+print_hit_cb(unsigned int start,
+             unsigned int end,
+             const char   *structure,
+             float        en,
+             void         *data);
 
 
 int
@@ -85,6 +89,7 @@ main(int  argc,
                                 split_contributions;
   long int                      first_alignment_number;
   float                         e_max;
+  double                        color_threshold, color_min_sat;
   vrna_md_t                     md;
   vrna_fold_compound_t          *fc;
   dataset_id                    id_control;
@@ -97,6 +102,8 @@ main(int  argc,
   ribo                    = 0;
   alnPS                   = 0;
   ssPS                    = 0;
+  color_threshold         = 2.;
+  color_min_sat           = 0.2;
   aln_out                 = 0;
   aln_columns             = 60;
   input_format_options    = VRNA_FILE_FORMAT_MSA_CLUSTAL; /* default to ClustalW format */
@@ -123,6 +130,9 @@ main(int  argc,
   if (RNALalifold_cmdline_parser(argc, argv, &args_info) != 0)
     exit(1);
 
+  /* prepare logging system and verbose mode */
+  ggo_log_settings(args_info, verbose);
+
   /* get basic set of model details */
   ggo_get_md_eval(args_info, md);
   ggo_get_md_fold(args_info, md);
@@ -131,8 +141,10 @@ main(int  argc,
   ggo_get_temperature(args_info, md.temperature);
 
   /* check dangle model */
-  if ((md.dangles < 0) || (md.dangles > 3)) {
-    vrna_message_warning("required dangle model not implemented, falling back to default dangles=2");
+  if (!((md.dangles == 0) || (md.dangles == 2))) {
+    vrna_log_warning(
+      "required dangle model %d not implemented, falling back to default dangles=2",
+      md.dangles);
     md.dangles = 2;
   }
 
@@ -166,6 +178,12 @@ main(int  argc,
   if (args_info.aln_EPS_ss_given)
     ssPS = 1;
 
+  if (args_info.color_threshold_given)
+    color_threshold = args_info.color_threshold_arg;
+
+  if (args_info.color_min_sat_given)
+    color_min_sat = args_info.color_min_sat_arg;
+
   if (args_info.aln_EPS_given)
     alnPS = 1;
 
@@ -176,9 +194,8 @@ main(int  argc,
     aln_out = 1;
     if (args_info.aln_stk_arg) {
       if (prefix) {
-        vrna_message_info(stdout,
-                          "multiple output prefixes detected, using \"%s\"",
-                          args_info.aln_stk_arg);
+        vrna_log_warning("multiple output prefixes detected, using \"%s\"",
+                         args_info.aln_stk_arg);
         free(prefix);
       }
 
@@ -196,12 +213,9 @@ main(int  argc,
     md.ribo     = ribo = 1;
   }
 
-  if (args_info.verbose_given)
-    verbose = 1;
-
   if (args_info.quiet_given) {
     if (verbose)
-      vrna_message_warning(
+      vrna_log_warning(
         "Can not be verbose and quiet at the same time! I keep on being chatty...");
     else
       quiet = 1;
@@ -210,7 +224,7 @@ main(int  argc,
   /* SHAPE reactivity data */
   if (args_info.shape_given) {
     if (verbose)
-      vrna_message_info(stderr, "SHAPE reactivity data correction activated");
+      vrna_log_info("SHAPE reactivity data correction activated");
 
     with_shapes             = 1;
     shape_files             = (char **)vrna_alloc(sizeof(char *) * (args_info.shape_given + 1));
@@ -236,8 +250,7 @@ main(int  argc,
       }
 
       if (verbose) {
-        vrna_message_info(stderr,
-                          "Using SHAPE reactivity data provided in file %s for sequence %d",
+        vrna_log_info("Using SHAPE reactivity data provided in file %s for sequence %d",
                           shape_files[s],
                           shape_file_association[s] + 1);
       }
@@ -256,8 +269,8 @@ main(int  argc,
     filename_in = strdup(args_info.inputs[0]);
     clust_file  = fopen(filename_in, "r");
     if (clust_file == NULL) {
-      vrna_message_warning("unable to open %s", filename_in);
-      vrna_message_error("Input file can't be read!");
+      vrna_log_error("unable to open input file \"%s\"", filename_in);
+      exit(EXIT_FAILURE);
     }
 
     /*
@@ -294,7 +307,7 @@ main(int  argc,
         break;
 
       default:
-        vrna_message_warning("Unknown input format specified");
+        vrna_log_warning("Unknown input format specified");
         break;
     }
   }
@@ -322,6 +335,8 @@ main(int  argc,
   if (args_info.split_contributions_given)
     split_contributions = 1;
 
+  ggo_geometry_settings(args_info, &md);
+
   /* free allocated memory of command line data structure */
   RNALalifold_cmdline_parser_free(&args_info);
 
@@ -370,10 +385,11 @@ main(int  argc,
           format = strdup("Unknown");
           break;
       }
-      vrna_message_error(
+      vrna_log_error(
         "Your input file is missing sequences! Either your file is empty, or not in %s format!",
         format);
       free(format);
+      exit(EXIT_FAILURE);
     }
 
     input_format_options = format_guess;
@@ -408,7 +424,8 @@ main(int  argc,
           break;
 
         default:
-          vrna_message_error("Which input format are you using?");
+          vrna_log_error("Which input format are you using?");
+          exit(EXIT_FAILURE);
           break;
       }
     }
@@ -463,6 +480,8 @@ main(int  argc,
     data.columns        = aln_columns;
     data.md             = &md;
     data.ss_eps         = ssPS;
+    data.color_threshold  = color_threshold;
+    data.color_min_sat    = color_min_sat;
     data.msa_eps        = alnPS;
     data.msa_stk        = aln_out;
     data.csv            = csv;
@@ -477,7 +496,7 @@ main(int  argc,
       for (s = 0; shape_file_association[s] != -1; s++);
 
       if ((s != n_seq) && (!quiet))
-        vrna_message_warning(
+        vrna_log_warning(
           "Number of sequences in alignment does not match number of provided SHAPE reactivity data files! ");
 
       shape_files             = (char **)vrna_realloc(shape_files, (n_seq + 1) * sizeof(char *));
@@ -538,10 +557,11 @@ main(int  argc,
         format = strdup("Unknown");
         break;
     }
-    vrna_message_error(
+    vrna_log_error(
       "Your input file is missing sequences! Either your file is empty, or not in %s format!",
       format);
     free(format);
+    exit(EXIT_FAILURE);
   }
 
   if (clust_file != stdin)
@@ -559,21 +579,24 @@ main(int  argc,
 
   free_id_data(id_control);
 
+  if (vrna_log_fp() != stderr)
+    fclose(vrna_log_fp());
+
   return EXIT_SUCCESS;
 }
 
 
 PRIVATE void
-print_hit_cb(int        start,
-             int        end,
-             const char *structure,
-             float      en,
-             void       *data)
+print_hit_cb(unsigned int start,
+             unsigned int end,
+             const char   *structure,
+             float        en,
+             void         *data)
 {
   char      **sub;
   char      *fname, *tmp_string;
   char      *id;
-  char      **A, *cons;
+  char      *cons;
   char      **names;
   char      **strings;
   char      **strings_orig;
@@ -583,6 +606,9 @@ print_hit_cb(int        start,
   vrna_md_t *md;
   int       with_ss, with_msa, with_stk, with_csv, with_mis, with_shapes, split_energies;
   float     threshold;
+  double    color_threshold, color_min_sat;
+  vrna_string_t *A;
+  short int     *pt;
 
   names           = ((hit_data *)data)->names;
   strings         = ((hit_data *)data)->strings;
@@ -591,6 +617,8 @@ print_hit_cb(int        start,
   columns         = ((hit_data *)data)->columns;
   md              = ((hit_data *)data)->md;
   with_ss         = ((hit_data *)data)->ss_eps;
+  color_threshold = ((hit_data *)data)->color_threshold;
+  color_min_sat   = ((hit_data *)data)->color_min_sat;
   with_msa        = ((hit_data *)data)->msa_eps;
   with_stk        = ((hit_data *)data)->msa_stk;
   with_csv        = ((hit_data *)data)->csv;
@@ -607,13 +635,20 @@ print_hit_cb(int        start,
   }
 
   if ((en / (float)(end - start + 1)) <= threshold) {
-    sub   = vrna_aln_slice((const char **)strings, (unsigned int)start, (unsigned int)end);
+    sub   = vrna_aln_slice((const char **)strings, start, end);
     cons  =
       (with_mis) ? vrna_aln_consensus_mis((const char **)sub,
                                           md) : vrna_aln_consensus_sequence((const char **)sub, md);
-    A = vrna_annotate_covar_db((const char **)sub, ss, md);
 
 
+    pt = vrna_ptable(ss);
+
+    A = vrna_annotate_covar_pt((const char **)sub,
+                               pt,
+                               md,
+                               color_threshold,
+                               color_min_sat);
+
     if (split_energies) {
       vrna_fold_compound_t  *sub_fc = vrna_fold_compound_comparative((const char **)sub,
                                                                      md,
@@ -623,14 +658,14 @@ print_hit_cb(int        start,
       if (with_shapes) {
         float cov_en = vrna_eval_covar_structure(sub_fc, ss);
         if (with_csv == 1) {
-          msg = vrna_strdup_printf(",%6.2f,%6.2f,%6.2f,%d,%d",
+          msg = vrna_strdup_printf(",%6.2f,%6.2f,%6.2f,%u,%u",
                                    real_en,
                                    -cov_en,
                                    en - real_en + cov_en,
                                    start,
                                    end);
         } else {
-          msg = vrna_strdup_printf(" (%6.2f = %6.2f + %6.2f + %6.2f) %4d - %4d",
+          msg = vrna_strdup_printf(" (%6.2f = %6.2f + %6.2f + %6.2f) %4u - %4u",
                                    en,
                                    real_en,
                                    -cov_en,
@@ -642,7 +677,7 @@ print_hit_cb(int        start,
         if (with_csv == 1) {
           msg = vrna_strdup_printf(",%6.2f,%6.2f,%d,%d", real_en, en - real_en, start, end);
         } else {
-          msg = vrna_strdup_printf(" (%6.2f = %6.2f + %6.2f) %4d - %4d",
+          msg = vrna_strdup_printf(" (%6.2f = %6.2f + %6.2f) %4u - %4u",
                                    en,
                                    real_en,
                                    en - real_en,
@@ -654,9 +689,9 @@ print_hit_cb(int        start,
       vrna_fold_compound_free(sub_fc);
     } else {
       if (with_csv == 1)
-        msg = vrna_strdup_printf(",%6.2f,%d,%d", en, start, end);
+        msg = vrna_strdup_printf(",%6.2f,%u,%u", en, start, end);
       else
-        msg = vrna_strdup_printf(" (%6.2f) %4d - %4d", en, start, end);
+        msg = vrna_strdup_printf(" (%6.2f) %4u - %4u", en, start, end);
     }
 
     print_structure(stdout, ss, msg);
@@ -664,9 +699,9 @@ print_hit_cb(int        start,
 
     if (with_ss) {
       if (prefix)
-        fname = vrna_strdup_printf("%s_ss_%d_%d.eps", prefix, start, end);
+        fname = vrna_strdup_printf("%s_ss_%u_%u.eps", prefix, start, end);
       else
-        fname = vrna_strdup_printf("ss_%d_%d.eps", start, end);
+        fname = vrna_strdup_printf("ss_%u_%u.eps", start, end);
 
       tmp_string = vrna_filename_sanitize(fname, "_");
       free(fname);
@@ -676,42 +711,47 @@ print_hit_cb(int        start,
       free(fname);
     }
 
-    free(A[0]);
-    free(A[1]);
+    free(pt);
+    vrna_string_free(A[0]);
+    vrna_string_free(A[1]);
     free(A);
     free(cons);
 
     if (with_msa) {
       if (prefix)
-        fname = vrna_strdup_printf("%s_aln_%d_%d.eps", prefix, start, end);
+        fname = vrna_strdup_printf("%s_aln_%u_%u.eps", prefix, start, end);
       else
-        fname = vrna_strdup_printf("aln_%d_%d.eps", start, end);
+        fname = vrna_strdup_printf("aln_%u_%u.eps", start, end);
 
       tmp_string = vrna_filename_sanitize(fname, "_");
 
-      vrna_file_PS_aln_slice(tmp_string,
-                             (const char **)strings_orig,
-                             (const char **)names,
-                             ss - start + 1,
-                             start,
-                             end,
-                             0,
-                             columns);
+      vrna_file_PS_aln_opt(tmp_string,
+                           (const char **)strings_orig,
+                           (const char **)names,
+                           ss - start + 1,
+                           (vrna_aln_opt_t){
+                            .start = start,
+                            .end = end,
+                            .offset = 0,
+                            .columns = columns,
+                            .color_threshold = color_threshold,
+                            .color_min_sat = color_min_sat
+                           });
       free(fname);
       free(tmp_string);
     }
 
     if (with_stk) {
       char          **sub_orig = vrna_aln_slice((const char **)strings_orig,
-                                                (unsigned int)start,
-                                                (unsigned int)end);
+                                                start,
+                                                end);
       unsigned int  options = VRNA_FILE_FORMAT_MSA_STOCKHOLM |
                               VRNA_FILE_FORMAT_MSA_APPEND;
       if (with_mis)
         options |= VRNA_FILE_FORMAT_MSA_MIS;
 
       if (prefix) {
-        id    = vrna_strdup_printf("%s_aln_%d_%d", prefix, start, end);
+        id    = vrna_strdup_printf("%s_aln_%u_%u", prefix, start, end);
         fname = vrna_strdup_printf("%s.stk", prefix);
 
         tmp_string = vrna_filename_sanitize(fname, "_");
@@ -727,7 +767,7 @@ print_hit_cb(int        start,
                             options);
         free(fname);
       } else {
-        id = vrna_strdup_printf("aln_%d_%d", start, end);
+        id = vrna_strdup_printf("aln_%u_%u", start, end);
         vrna_file_msa_write("RNALalifold_results.stk",
                             (const char **)names,
                             (const char **)sub_orig,
