@@ -20,18 +20,23 @@
 #include "ViennaRNA/datastructures/basic.h"
 #include "ViennaRNA/utils/basic.h"
 #include "ViennaRNA/utils/strings.h"
+#include "ViennaRNA/utils/log.h"
 #include "ViennaRNA/params/io.h"
-#include "ViennaRNA/mfe.h"
+#include "ViennaRNA/mfe/local.h"
 #include "ViennaRNA/Lfold.h"
 #include "ViennaRNA/io/file_formats.h"
 #include "ViennaRNA/io/utils.h"
-#include "ViennaRNA/commands.h"
-#include "ViennaRNA/constraints/SHAPE.h"
-#include "RNALfold_cmdl.h"
-#include "gengetopt_helper.h"
-#include "input_id_helpers.h"
+#include "ViennaRNA/io/commands.h"
+#include "ViennaRNA/probing/SHAPE.h"
+#include "ViennaRNA/constraints/soft_special.h"
 
-#include "ViennaRNA/color_output.inc"
+#include "RNALfold_cmdl.h"
+#include "gengetopt_helpers.h"
+#include "input_id_helpers.h"
+#include "modified_bases_helpers.h"
+#include "probing_data_helpers.h"
+
+#include "ViennaRNA/intern/color_output.h"
 
 typedef struct {
   FILE  *output;
@@ -41,22 +46,22 @@ typedef struct {
 
 #ifdef VRNA_WITH_SVM
 PRIVATE void
-default_callback_z(int        start,
-                   int        end,
-                   const char *structure,
-                   float      en,
-                   float      zscore,
-                   void       *data);
+default_callback_z(unsigned int start,
+                   unsigned int end,
+                   const char   *structure,
+                   float        en,
+                   float        zscore,
+                   void         *data);
 
 
 #endif
 
 PRIVATE void
-default_callback(int        start,
-                 int        end,
-                 const char *structure,
-                 float      en,
-                 void       *data);
+default_callback(unsigned int start,
+                 unsigned int end,
+                 const char   *structure,
+                 float        en,
+                 void         *data);
 
 
 int
@@ -70,31 +75,39 @@ main(int  argc,
                               *shape_file, *shape_method, *shape_conversion;
   unsigned int                rec_type, read_opt;
   int                         length, istty, noconv, maxdist, zsc, tofile, filename_full,
-                              with_shapes, verbose;
+                              with_shapes, verbose, backtrack, zsc_pre, zsc_subsumed;
   double                      min_en, min_z;
+  long int                    file_pos_start;
   vrna_md_t                   md;
   vrna_cmd_t                  commands;
   dataset_id                  id_control;
+  vrna_sc_mod_param_t         *mod_params;
 
-  ParamFile     = ns_bases = NULL;
-  do_backtrack  = 1;
-  noconv        = 0;
-  dangles       = 2;
-  maxdist       = 150;
-  zsc           = 0;
-  min_z         = -2.0;
-  gquad         = 0;
-  rec_type      = read_opt = 0;
-  rec_id        = rec_sequence = orig_sequence = NULL;
-  rec_rest      = NULL;
-  outfile       = NULL;
-  infile        = NULL;
-  input         = NULL;
-  output        = NULL;
-  tofile        = 0;
-  filename_full = 0;
-  command_file  = NULL;
-  commands      = NULL;
+  ParamFile           = ns_bases = NULL;
+  do_backtrack        = 1;
+  noconv              = 0;
+  backtrack           = 0;
+  dangles             = 2;
+  maxdist             = 150;
+  zsc                 = 0;
+  zsc_pre             = 0;
+  zsc_subsumed        = 0;
+  min_z               = -2.0;
+  gquad               = 0;
+  rec_type            = read_opt = 0;
+  rec_id              = rec_sequence = orig_sequence = NULL;
+  rec_rest            = NULL;
+  outfile             = NULL;
+  infile              = NULL;
+  input               = NULL;
+  output              = NULL;
+  tofile              = 0;
+  filename_full       = 0;
+  command_file        = NULL;
+  commands            = NULL;
+  file_pos_start      = -1;
+  verbose             = 0;
+  mod_params          = NULL;
 
   /* apply default model details */
   vrna_md_set_default(&md);
@@ -107,6 +120,9 @@ main(int  argc,
   if (RNALfold_cmdline_parser(argc, argv, &args_info) != 0)
     exit(1);
 
+  /* prepare logging system and verbose mode */
+  ggo_log_settings(args_info, verbose);
+
   /* parse options for ID manipulation */
   ggo_get_id_control(args_info, id_control, "Sequence", "sequence", "_", 4, 1);
 
@@ -121,12 +137,24 @@ main(int  argc,
   /* set dangle model */
   if (args_info.dangles_given) {
     if ((args_info.dangles_arg < 0) || (args_info.dangles_arg > 3))
-      vrna_message_warning(
+      vrna_log_warning(
         "required dangle model not implemented, falling back to default dangles=2");
     else
       md.dangles = dangles = args_info.dangles_arg;
   }
 
+  /* do not allow weak pairs */
+  if (args_info.backtrack_global_given) {
+    if (dangles % 2) {
+      vrna_log_warning(
+        "Global mfE structure backtracking not available for odd dangle models yet!"
+        " Deactivating global backtracing now...");
+      backtrack = 0;
+    } else {
+      backtrack = tofile = 1;
+    }
+  }
+
   /* do not allow weak pairs */
   if (args_info.noLP_given)
     md.noLP = noLonelyPairs = 1;
@@ -139,6 +167,10 @@ main(int  argc,
   if (args_info.noClosingGU_given)
     md.noGUclosure = no_closingGU = 1;
 
+  /* set salt concentration */
+  if (args_info.salt_given)
+    md.salt = args_info.salt_arg;
+
   /* do not convert DNA nucleotide "T" to appropriate RNA "U" */
   if (args_info.noconv_given)
     noconv = 1;
@@ -148,8 +180,7 @@ main(int  argc,
     md.energy_set = energy_set = args_info.energyModel_arg;
 
   /* take another energy parameter set */
-  if (args_info.paramFile_given)
-    ParamFile = strdup(args_info.paramFile_arg);
+  ggo_get_read_paramFile(args_info, &md);
 
   /* Allow other pairs in addition to the usual AU,GC,and GU pairs */
   if (args_info.nsp_given)
@@ -165,8 +196,16 @@ main(int  argc,
     if (args_info.zscore_arg != -2)
       min_z = args_info.zscore_arg;
 
+    if ((args_info.zscore_pre_filter_given) ||
+        (backtrack)) /* global backtracing implies hard z-score filter! */
+      zsc_pre = 1;
+
+    if (args_info.zscore_report_subsumed_given)
+      zsc_subsumed = 1;
+
 #else
-    vrna_message_error("\'z\' option is available only if compiled with SVM support!");
+    vrna_log_error("\'z\' option is available only if compiled with SVM support!");
+    exit(EXIT_FAILURE);
 #endif
   }
 
@@ -174,9 +213,6 @@ main(int  argc,
   if (args_info.gquad_given)
     md.gquad = gquad = 1;
 
-  if (args_info.verbose_given)
-    verbose = 1;
-
   /* SHAPE reactivity data */
   ggo_get_SHAPE(args_info, with_shapes, shape_file, shape_method, shape_conversion);
 
@@ -215,6 +251,12 @@ main(int  argc,
     exit(EXIT_FAILURE);
   }
 
+  ggo_get_modified_base_settings(args_info,
+                                 mod_params,
+                                 &(md));
+
+  ggo_geometry_settings(args_info, &md);
+
   /* free allocated memory of command line data structure */
   RNALfold_cmdline_parser_free(&args_info);
 
@@ -228,18 +270,14 @@ main(int  argc,
 
   if (infile) {
     input = fopen((const char *)infile, "r");
-    if (!input)
-      vrna_message_error("Could not read input file");
+    if (!input) {
+      vrna_log_error("Could not read input file");
+      exit(EXIT_FAILURE);
+    }
   } else {
     input = stdin;
   }
 
-  if (ParamFile != NULL) {
-    if (!strcmp(ParamFile, "DNA"))
-        vrna_params_load_DNA_Mathews2004();
-    else
-      vrna_params_load(ParamFile, VRNA_PARAMETER_FORMAT_DEFAULT);
-  }
 
   if (command_file != NULL)
     commands = vrna_file_commands_read(command_file, VRNA_CMD_PARSE_HC | VRNA_CMD_PARSE_SC);
@@ -267,9 +305,11 @@ main(int  argc,
      # init everything according to the data we've read
      ########################################################
      */
-    char  *SEQ_ID       = NULL;
-    char  *v_file_name  = NULL;
-    char  *tmp_string   = NULL;
+    size_t  **mod_positions, mod_param_sets;
+    char    *SEQ_ID       = NULL;
+    char    *v_file_name  = NULL;
+    char    *tmp_string   = NULL;
+
     /*
      ########################################################
      # init everything according to the data we've read
@@ -295,12 +335,19 @@ main(int  argc,
       free(v_file_name);
       v_file_name = tmp_string;
 
-      if (infile && !strcmp(infile, v_file_name))
-        vrna_message_error("Input and output file names are identical");
+      if (infile && !strcmp(infile, v_file_name)) {
+        vrna_log_error("Input and output file names are identical");
+        exit(EXIT_FAILURE);
+      }
 
       output = fopen((const char *)v_file_name, "a");
-      if (!output)
-        vrna_message_error("Failed to open file for writing");
+
+      if (!output) {
+        vrna_log_error("Failed to open file for writing");
+        exit(EXIT_FAILURE);
+      }
+
+      file_pos_start = ftell(output);
     } else {
       output = stdout;
     }
@@ -310,17 +357,23 @@ main(int  argc,
 
     length = (int)strlen(rec_sequence);
 
+    mod_positions = mod_positions_seq_prepare(rec_sequence,
+                                              mod_params,
+                                              verbose,
+                                              &mod_param_sets);
+
     /* convert DNA alphabet to RNA if not explicitely switched off */
     if (!noconv)
       vrna_seq_toRNA(rec_sequence);
 
     /* store case-unmodified sequence */
     orig_sequence = strdup(rec_sequence);
+
     /* convert sequence to uppercase letters only */
     vrna_seq_toupper(rec_sequence);
 
     if (!tofile && istty)
-      vrna_message_info(output, "length = %d", length);
+      printf("length = %d", length);
 
     /*
      ########################################################
@@ -345,6 +398,27 @@ main(int  argc,
                                  VRNA_OPTION_WINDOW);
     }
 
+    /* apply modified base support if requested */
+    mod_bases_apply(vc,
+                    mod_param_sets,
+                    mod_positions,
+                    mod_params);
+
+#ifdef VRNA_WITH_SVM
+    if (zsc) {
+      unsigned int zsc_options = VRNA_ZSCORE_FILTER_ON;
+
+      if (zsc_pre)
+        zsc_options |= VRNA_ZSCORE_PRE_FILTER;
+
+      if (zsc_subsumed)
+        zsc_options |= VRNA_ZSCORE_REPORT_SUBSUMED;
+
+      vrna_zsc_filter_init(vc, min_z, zsc_options);
+    }
+
+#endif
+
     hit_data data;
     data.output       = output;
     data.dangle_model = md.dangles;
@@ -359,13 +433,27 @@ main(int  argc,
 #endif
     fprintf(output, "%s\n", orig_sequence);
 
-    char *msg = NULL;
-    if (!tofile && istty)
-      msg = vrna_strdup_printf(" minimum free energy = %6.2f kcal/mol", min_en);
-    else
-      msg = vrna_strdup_printf(" (%6.2f)", min_en);
+    char  *msg            = NULL;
+    char  *mfe_structure  = NULL;
 
-    print_structure(output, NULL, msg);
+    if (output)
+      (void)fflush(output);
+
+    if (backtrack) {
+      if (vrna_backtrack_window(vc,
+                                (const char *)v_file_name,
+                                file_pos_start,
+                                &mfe_structure,
+                                min_en))
+        msg = vrna_strdup_printf(" (%6.2f)", min_en);
+    } else {
+      if (!tofile && istty)
+        msg = vrna_strdup_printf(" minimum free energy = %6.2f kcal/mol", min_en);
+      else
+        msg = vrna_strdup_printf(" (%6.2f)", min_en);
+    }
+
+    print_structure(output, mfe_structure, msg);
     free(msg);
 
     if (output)
@@ -403,18 +491,28 @@ main(int  argc,
   free(command_file);
   vrna_commands_free(commands);
 
+  if (mod_params) {
+    for (vrna_sc_mod_param_t *ptr = mod_params; *ptr != NULL; ptr++)
+      vrna_sc_mod_parameters_free(*ptr);
+
+    free(mod_params);
+  }
+
   free_id_data(id_control);
 
+  if (vrna_log_fp() != stderr)
+    fclose(vrna_log_fp());
+
   return EXIT_SUCCESS;
 }
 
 
 PRIVATE void
-default_callback(int        start,
-                 int        end,
-                 const char *structure,
-                 float      en,
-                 void       *data)
+default_callback(unsigned int start,
+                 unsigned int end,
+                 const char   *structure,
+                 float        en,
+                 void         *data)
 {
   FILE  *output       = ((hit_data *)data)->output;
   int   dangle_model  = ((hit_data *)data)->dangle_model;
@@ -422,12 +520,12 @@ default_callback(int        start,
   char  *msg          = NULL;
 
   if ((dangle_model == 2) && (start > 1)) {
-    msg       = vrna_strdup_printf(" (%6.2f) %4d", en, start - 1);
+    msg       = vrna_strdup_printf(" (%6.2f) %4u", en, start - 1);
     struct_d2 = vrna_strdup_printf(".%s", structure);
     print_structure(output, struct_d2, msg);
     free(struct_d2);
   } else {
-    msg = vrna_strdup_printf(" (%6.2f) %4d", en, start);
+    msg = vrna_strdup_printf(" (%6.2f) %4u", en, start);
     print_structure(output, structure, msg);
   }
 
@@ -437,12 +535,12 @@ default_callback(int        start,
 
 #ifdef VRNA_WITH_SVM
 PRIVATE void
-default_callback_z(int        start,
-                   int        end,
-                   const char *structure,
-                   float      en,
-                   float      zscore,
-                   void       *data)
+default_callback_z(unsigned int start,
+                   unsigned int end,
+                   const char   *structure,
+                   float        en,
+                   float        zscore,
+                   void         *data)
 {
   FILE  *output       = ((hit_data *)data)->output;
   int   dangle_model  = ((hit_data *)data)->dangle_model;
@@ -450,12 +548,12 @@ default_callback_z(int        start,
   char  *msg          = NULL;
 
   if ((dangle_model == 2) && (start > 1)) {
-    msg       = vrna_strdup_printf(" (%6.2f) %4d z= %.3f", en, start - 1, zscore);
+    msg       = vrna_strdup_printf(" (%6.2f) %4u z= %.3f", en, start - 1, zscore);
     struct_d2 = vrna_strdup_printf(".%s", structure);
     print_structure(output, struct_d2, msg);
     free(struct_d2);
   } else {
-    msg = vrna_strdup_printf(" (%6.2f) %4d z= %.3f", en, start, zscore);
+    msg = vrna_strdup_printf(" (%6.2f) %4u z= %.3f", en, start, zscore);
     print_structure(output, structure, msg);
   }
 
