@@ -18,31 +18,36 @@
 #include "ViennaRNA/plotting/structures.h"
 #include "ViennaRNA/fold_vars.h"
 #include "ViennaRNA/params/basic.h"
-#include "ViennaRNA/commands.h"
+#include "ViennaRNA/io/commands.h"
+#include "ViennaRNA/combinatorics/basic.h"
 #include "ViennaRNA/constraints/basic.h"
 #include "ViennaRNA/constraints/hard.h"
 #include "ViennaRNA/constraints/soft.h"
-#include "ViennaRNA/constraints/SHAPE.h"
+#include "ViennaRNA/probing/SHAPE.h"
+#include "ViennaRNA/constraints/soft_special.h"
 #include "ViennaRNA/io/file_formats.h"
 #include "ViennaRNA/io/utils.h"
 #include "ViennaRNA/cofold.h"
 #include "ViennaRNA/fold.h"
 #include "ViennaRNA/part_func_co.h"
-#include "ViennaRNA/part_func.h"
-#include "ViennaRNA/centroid.h"
-#include "ViennaRNA/MEA.h"
+#include "ViennaRNA/partfunc/global.h"
+#include "ViennaRNA/structures/centroid.h"
+#include "ViennaRNA/structures/mea.h"
 #include "ViennaRNA/utils/basic.h"
 #include "ViennaRNA/utils/strings.h"
-#include "ViennaRNA/utils/structures.h"
+#include "ViennaRNA/structures/pairtable.h"
+#include "ViennaRNA/utils/log.h"
 #include "ViennaRNA/params/io.h"
 #include "ViennaRNA/datastructures/char_stream.h"
 #include "ViennaRNA/datastructures/stream_output.h"
 
 #include "RNAcofold_cmdl.h"
-#include "gengetopt_helper.h"
+#include "gengetopt_helpers.h"
 #include "input_id_helpers.h"
-#include "ViennaRNA/color_output.inc"
+#include "modified_bases_helpers.h"
+#include "ViennaRNA/intern/color_output.h"
 #include "parallel_helpers.h"
+#include "probing_data_helpers.h"
 
 
 struct options {
@@ -75,6 +80,8 @@ struct options {
   char            *shape_method;
   char            *shape_conversion;
 
+  vrna_sc_mod_param_t *mod_params;
+
   int             csv_output;
   int             csv_header;
   char            csv_output_delim;
@@ -199,6 +206,8 @@ init_default_options(struct options *opt)
   opt->shape_method     = NULL;
   opt->shape_conversion = NULL;
 
+  opt->mod_params         = NULL;
+
   opt->csv_output       = 0;    /* flag indicating whether we produce one-line outputs, a.k.a. CSV */
   opt->csv_header       = 1;    /* print header for one-line output */
   opt->csv_output_delim = ',';  /* delimiting character for one-line output */
@@ -268,6 +277,9 @@ main(int  argc,
   if (RNAcofold_cmdline_parser(argc, argv, &args_info) != 0)
     exit(1);
 
+  /* prepare logging system and verbose mode */
+  ggo_log_settings(args_info, opt.verbose);
+
   /* get basic set of model details */
   ggo_get_md_eval(args_info, opt.md);
   ggo_get_md_fold(args_info, opt.md);
@@ -278,7 +290,7 @@ main(int  argc,
 
   /* check dangle model */
   if ((opt.md.dangles < 0) || (opt.md.dangles > 3)) {
-    vrna_message_warning("required dangle model not implemented, falling back to default dangles=2");
+    vrna_log_warning("required dangle model not implemented, falling back to default dangles=2");
     opt.md.dangles = dangles = 2;
   }
 
@@ -318,6 +330,10 @@ main(int  argc,
     opt.concentration_file  = strdup(args_info.concfile_arg);
     opt.doC                 = opt.doT = opt.pf = 1;
   }
+  
+  /* Salt duplex init */
+  if (args_info.saltInit_given)
+    opt.md.saltDPXInit = args_info.saltInit_arg;
 
   /* partition function settings */
   if (args_info.partfunc_given) {
@@ -358,13 +374,14 @@ main(int  argc,
       opt.md.compute_bpp = 1;
   }
 
-  if (args_info.verbose_given)
-    opt.verbose = 1;
-
   if (args_info.commands_given)
     opt.commands = vrna_file_commands_read(args_info.commands_arg,
                                            VRNA_CMD_PARSE_HC | VRNA_CMD_PARSE_SC);
 
+  ggo_get_modified_base_settings(args_info,
+                                 opt.mod_params,
+                                 &(opt.md));
+
   /* filename sanitize delimiter */
   if (args_info.filename_delim_given)
     opt.filename_delim = strdup(args_info.filename_delim_arg);
@@ -394,7 +411,7 @@ main(int  argc,
         opt.csv_output = 0;
         break;
       default:
-        vrna_message_warning("unknown output format \"%c\", using defaults!",
+        vrna_log_warning("unknown output format \"%c\", using defaults!",
                              *(args_info.output_format_arg));
         break;
     }
@@ -404,7 +421,7 @@ main(int  argc,
   if (args_info.csv_delim_given) {
     opt.csv_output_delim = *(args_info.csv_delim_arg);
     if (!opt.csv_output_delim) {
-      vrna_message_warning("Delimiting character for One-Line output is missing, using defaults!");
+      vrna_log_warning("Delimiting character for One-Line output is missing, using defaults!");
       opt.csv_output_delim = ',';
     }
   }
@@ -422,7 +439,7 @@ main(int  argc,
       if (num_proc_cores(&proc_cores, &proc_cores_conf)) {
         opt.jobs = MIN2(thread_max, proc_cores_conf);
       } else {
-        vrna_message_warning("Could not determine number of available processor cores!\n"
+        vrna_log_warning("Could not determine number of available processor cores!\n"
                              "Defaulting to serial computation");
         opt.jobs = 1;
       }
@@ -432,7 +449,7 @@ main(int  argc,
 
     opt.jobs = MAX2(1, opt.jobs);
 #else
-    vrna_message_warning(
+    vrna_log_warning(
       "This version of RNAfold has been built without parallel input processing capabilities");
 #endif
 
@@ -440,6 +457,8 @@ main(int  argc,
       opt.keep_order = 0;
   }
 
+  ggo_geometry_settings(args_info, &(opt.md));
+
   input_files = collect_unnamed_options(&args_info, &num_input);
 
   /* free allocated memory of command line data structure */
@@ -450,15 +469,17 @@ main(int  argc,
    # begin initializing
    #############################################
    */
-  if (opt.pf && opt.md.gquad)
-    vrna_message_error(
+  if (opt.pf && opt.md.gquad) {
+    vrna_log_error(
       "G-Quadruplex support is currently not available for partition function computations");
+    exit(EXIT_FAILURE);
+  }
 
   if ((opt.csv_output) && (opt.csv_header))
     write_csv_header(stdout, &opt);
 
   if ((opt.verbose) && (opt.jobs > 1))
-    vrna_message_info(stderr, "Preparing %d parallel computation slots", opt.jobs);
+    vrna_log_info("Preparing %d parallel computation slots", opt.jobs);
 
   if (opt.keep_order)
     opt.output_queue = vrna_ostream_init(&flush_cstr_callback, NULL);
@@ -476,13 +497,14 @@ main(int  argc,
       if (!skip) {
         FILE *input_stream = fopen((const char *)input_files[i], "r");
 
-        if (!input_stream)
-          vrna_message_error("Unable to open %d. input file \"%s\" for reading", i + 1,
+        if (!input_stream) {
+          vrna_log_error("Unable to open %d. input file \"%s\" for reading", i + 1,
                              input_files[i]);
+          exit(EXIT_FAILURE);
+        }
 
         if (opt.verbose) {
-          vrna_message_info(stderr,
-                            "Processing %d. input file \"%s\"",
+          vrna_log_info("Processing %d. input file \"%s\"",
                             i + 1,
                             input_files[i]);
         }
@@ -517,8 +539,18 @@ main(int  argc,
   vrna_commands_free(opt.commands);
   free(opt.concentration_file);
 
+  if (opt.mod_params) {
+    for (vrna_sc_mod_param_t *ptr = opt.mod_params; *ptr != NULL; ptr++)
+      vrna_sc_mod_parameters_free(*ptr);
+
+    free(opt.mod_params);
+  }
+
   free_id_data(opt.id_control);
 
+  if (vrna_log_fp() != stderr)
+    fclose(vrna_log_fp());
+
   return EXIT_SUCCESS;
 }
 
@@ -645,6 +677,8 @@ process_record(struct record_data *record)
   vrna_ep_t             *prAB, *prAA, *prBB, *prA, *prB, *mfAB, *mfAA, *mfBB, *mfA, *mfB;
   struct options        *opt;
   struct output_stream  *o_stream;
+  size_t                **mod_positions;
+  size_t                mod_param_sets;
 
   mfAB            = mfAA = mfBB = mfA = mfB = NULL;
   prAB            = prAA = prBB = prA = prB = NULL;
@@ -654,6 +688,11 @@ process_record(struct record_data *record)
   sequence        = strdup(record->sequence);
   rec_rest        = record->rest;
 
+  mod_positions   = mod_positions_seq_prepare(sequence,
+                                              opt->mod_params,
+                                              opt->verbose,
+                                              &mod_param_sets);
+
   /* convert DNA alphabet to RNA if not explicitely switched off */
   if (!opt->noconv) {
     vrna_seq_toRNA(sequence);
@@ -666,29 +705,42 @@ process_record(struct record_data *record)
   vrna_fold_compound_t *vc = vrna_fold_compound(sequence,
                                                 &(opt->md),
                                                 VRNA_OPTION_DEFAULT | VRNA_OPTION_HYBRID);
+
+  if (!vc) {
+    vrna_log_warning("Skipping computations for \"%s\"",
+                         (record->id) ? record->id : "identifier unavailable");
+    return;
+  }
+
   n = vc->length;
 
+  if (vc->strands > 2) {
+    vrna_log_error("More than one strand delimiter in input!");
+    exit(EXIT_FAILURE);
+  }
+
   /* retrieve string stream bound to stdout, 6*length should be enough memory to start with */
   o_stream->data = vrna_cstr(6 * n, stdout);
   /* retrieve string stream bound to stderr for any info messages */
   o_stream->err = vrna_cstr(n, stderr);
 
   if (record->tty) {
-    if (vc->cutpoint == -1) {
-      vrna_message_info(stdout, "length = %u", n);
+    if (vc->strands == 1) {
+      printf("length = %u", n);
     } else {
-      vrna_message_info(stdout,
-                        "length1 = %d\nlength2 = %d",
-                        vc->cutpoint - 1,
-                        (int)n - vc->cutpoint + 1);
+      printf("length1 = %d\nlength2 = %d",
+             vc->nucleotides[0].length,
+             vc->nucleotides[1].length);
     }
   }
 
-  if (vc->cutpoint == vc->length / 2 + 1) {
-    if (!strncmp(vc->sequence, vc->sequence + vc->cutpoint - 1, vc->cutpoint - 1)) {
+  if (vc->strands > 1) {
+    unsigned int r = vrna_rotational_symmetry(vc->sequence);
+    if (r > 1) {
       vrna_cstr_message_warning(o_stream->err,
-                                "Both input strands are identical, thus inducing rotationally symmetry! "
-                                "Symmetry correction might be required to compute actual MFE!");
+                                "The input sequences show a rotational symmetry of %u! "
+                                "Symmetry correction might be required to compute actual MFE!",
+                                r);
     }
   }
 
@@ -705,21 +757,57 @@ process_record(struct record_data *record)
       int           cp        = -1;
 
       cstruc  = vrna_extract_record_rest_structure((const char **)rec_rest, 0, coptions);
-      cstruc  = vrna_cut_point_remove(cstruc, &cp);
-      if (vc->cutpoint != cp) {
-        vrna_message_error("Sequence and Structure have different cut points.\n"
-                           "sequence: %d, structure: %d",
-                           vc->cutpoint, cp);
+
+      char **structures = vrna_strsplit(cstruc, "&");
+      for (unsigned int a = 0; a < vc->strands; a++) {
+        if (!structures[a]) {
+          vrna_log_error("Sequence and Structure have different number of strand delimiters");
+          exit(EXIT_FAILURE);
+        }
+
+        unsigned int l = strlen(structures[a]);
+
+        cl += l;
+
+        switch (vc->type) {
+          case VRNA_FC_TYPE_SINGLE:
+            if (vc->nucleotides[a].length != l) {
+              vrna_log_error(
+                "Structure and sequence part of strand %u differ in length (%u vs. %u)",
+                a,
+                l,
+                vc->nucleotides[a].length);
+              exit(EXIT_FAILURE);
+            }
+
+            break;
+
+          case VRNA_FC_TYPE_COMPARATIVE:
+            if (vc->alignment[a].sequences[0].length != l) {
+              vrna_log_error(
+                "Structure and sequence part of strand %u differ in length (%u vs. %u)",
+                a,
+                l,
+                vc->alignment[a].sequences[0].length);
+              exit(EXIT_FAILURE);
+            }
+            break;
+        }
       }
 
-      cl = (cstruc) ? (int)strlen(cstruc) : 0;
+      for (unsigned int a = 0; a < vc->strands; a++)
+        free(structures[a]);
 
-      if (cl == 0)
-        vrna_message_warning("Structure constraint is missing");
-      else if (cl < n)
-        vrna_message_warning("Structure constraint is shorter than sequence");
-      else if (cl > n)
-        vrna_message_error("Structure constraint is too long");
+      free(structures);
+
+      if (cl == 0) {
+        vrna_log_warning("Structure constraint is missing");
+      } else if (cl < n) {
+        vrna_log_warning("Structure constraint is shorter than sequence");
+      } else if (cl > n) {
+        vrna_log_error("Structure constraint is too long");
+        exit(EXIT_FAILURE);
+      }
 
       if (cstruc) {
         unsigned int constraint_options = VRNA_CONSTRAINT_DB_DEFAULT;
@@ -749,12 +837,20 @@ process_record(struct record_data *record)
   if (opt->commands)
     vrna_commands_apply(vc, opt->commands, VRNA_CMD_PARSE_HC | VRNA_CMD_PARSE_SC);
 
+  /* apply modified base support if requested */
+  mod_bases_apply(vc,
+                  mod_param_sets,
+                  mod_positions,
+                  opt->mod_params);
+
   if (opt->doC) {
     if (opt->concentration_file) {
       /* read from file */
       FILE *fp = fopen(opt->concentration_file, "r");
-      if (fp == NULL)
-        vrna_message_error("could not open concentration file %s", opt->concentration_file);
+      if (fp == NULL) {
+        vrna_log_error("could not open concentration file %s", opt->concentration_file);
+        exit(EXIT_FAILURE);
+      }
 
       concentrations = read_concentrations(fp);
       fclose(fp);
@@ -777,7 +873,7 @@ process_record(struct record_data *record)
   /* check whether the constraint allows for any solution */
   if ((fold_constrained) || (opt->commands)) {
     if (min_en == (double)(INF / 100.)) {
-      vrna_message_error(
+      vrna_log_error(
         "Supplied structure constraints create empty solution set for sequence:\n%s",
         record->sequence);
       exit(EXIT_FAILURE);
@@ -785,7 +881,18 @@ process_record(struct record_data *record)
   }
 
   {
-    char *pstruct = vrna_cut_point_insert(mfe_structure, vc->cutpoint);
+    char  *pstruct    = NULL;
+    char  *tmp_struct = strdup(mfe_structure);
+
+    if (vc->strands == 1) {
+      pstruct = tmp_struct;
+    } else {
+      for (unsigned int i = 1; i < vc->strands; i++) {
+        pstruct = vrna_cut_point_insert(tmp_struct, (int)vc->strand_start[i] + (i - 1));
+        free(tmp_struct);
+        tmp_struct = pstruct;
+      }
+    }
 
     if (opt->csv_output) {
       vrna_cstr_printf(o_stream->data,
@@ -853,11 +960,27 @@ process_record(struct record_data *record)
     /* compute partition function */
     AB = AA = BB = vrna_pf_dimer(vc, pairing_propensity);
 
+    if(AB.FcAB == 999) {
+      vrna_log_error("underflow in calculating the free energy of connected ensamble FcAB\n"
+        "use smaller pf_scale");
+    }
+
     if (opt->md.compute_bpp) {
-      char *costruc;
+      char *costruc = NULL;
       prAB = vrna_plist_from_probs(vc, opt->bppmThreshold);
 
-      costruc = vrna_cut_point_insert(pairing_propensity, vc->cutpoint);
+      char  *tmp_struct = strdup(pairing_propensity);
+
+      if (vc->strands == 1) {
+        costruc = tmp_struct;
+      } else {
+        for (unsigned int i = 1; i < vc->strands; i++) {
+          costruc = vrna_cut_point_insert(tmp_struct, (int)vc->strand_start[i] + (i - 1));
+          free(tmp_struct);
+          tmp_struct = costruc;
+        }
+      }
+
       if (opt->csv_output) {
         vrna_cstr_printf(o_stream->data,
                          "%c"
@@ -900,9 +1023,26 @@ process_record(struct record_data *record)
                   o_stream->data);
     }
 
+    if (opt->csv_output) {
+      vrna_cstr_printf(o_stream->data,
+                       "%c"
+                       "%g%c"   /* probability of MFE structure */
+                       "%6.2f", /* delta G binding */
+                       opt->csv_output_delim,
+                       exp((AB.FAB - min_en) / kT), opt->csv_output_delim,
+                       AB.FcAB - AB.FA - AB.FB);
+    } else {
+      vrna_cstr_printf_structure(o_stream->data,
+                                 NULL,
+                                 " frequency of mfe structure in ensemble %g"
+                                 "; delta G binding=%6.2f",
+                                 exp((AB.FAB - min_en) / kT),
+                                 AB.FcAB - AB.FA - AB.FB);
+    }
+
     if (opt->doT) {
-      if (vc->cutpoint <= 0) {
-        vrna_message_warning(
+      if (vc->strands < 2) {
+        vrna_log_warning(
           "Sorry, i cannot do that with only one molecule, please give me two or leave it");
         free(mfAB);
         free(prAB);
@@ -910,6 +1050,64 @@ process_record(struct record_data *record)
         goto cleanup_record;
       }
 
+#if 1
+      Alength       = vc->nucleotides[0].length;                        /* length of first molecule */
+      Blength       = vc->nucleotides[1].length;                        /* length of 2nd molecule   */
+      orig_Astring  = (char *)vrna_alloc(sizeof(char) * (Alength + 1)); /*Sequence of first molecule*/
+      orig_Bstring  = (char *)vrna_alloc(sizeof(char) * (Blength + 1)); /*Sequence of second molecule*/
+      strncat(orig_Astring, record->sequence, Alength);
+      strncat(orig_Bstring, record->sequence + Alength + 1, Blength);
+
+      char                  *seq_AA = vrna_strdup_printf("%s&%s",
+                                                         vc->nucleotides[0].string,
+                                                         vc->nucleotides[0].string);
+
+      char                  *ss_AA  = (char *)vrna_alloc(sizeof(char) * (2 * Alength + 1));
+      vrna_fold_compound_t  *fc_AA  = vrna_fold_compound(seq_AA, &(opt->md), VRNA_OPTION_DEFAULT);
+
+      double                mfe_AA = vrna_mfe(fc_AA, ss_AA);
+      vrna_mx_mfe_free(fc_AA);
+
+      vrna_exp_params_rescale(fc_AA, &mfe_AA);
+
+      AA = vrna_pf_dimer(fc_AA, NULL);
+
+      mfAA  = vrna_plist(ss_AA, 0.95);
+      prAA  = vrna_plist_from_probs(fc_AA, opt->bppmThreshold);
+
+      free(seq_AA);
+      free(ss_AA);
+      vrna_fold_compound_free(fc_AA);
+
+      char                  *seq_BB = vrna_strdup_printf("%s&%s",
+                                                         vc->nucleotides[1].string,
+                                                         vc->nucleotides[1].string);
+
+      char                  *ss_BB  = (char *)vrna_alloc(sizeof(char) * (2 * Blength + 1));
+      vrna_fold_compound_t  *fc_BB  = vrna_fold_compound(seq_BB, &(opt->md), VRNA_OPTION_DEFAULT);
+
+      double                mfe_BB = vrna_mfe(fc_BB, ss_BB);
+      vrna_mx_mfe_free(fc_BB);
+
+      vrna_exp_params_rescale(fc_BB, &mfe_BB);
+
+      BB = vrna_pf_dimer(fc_BB, NULL);
+
+      mfBB  = vrna_plist(ss_BB, 0.95);
+      prBB  = vrna_plist_from_probs(fc_BB, opt->bppmThreshold);
+
+      free(seq_BB);
+      free(ss_BB);
+      vrna_fold_compound_free(fc_BB);
+
+      /* compute A monomer */
+      do_partfunc(vc->nucleotides[0].string, Alength, 1, &prA, &mfA, kT, opt);
+
+      /* compute B monomer */
+      do_partfunc(vc->nucleotides[1].string, Blength, 1, &prB, &mfB, kT, opt);
+
+
+#else
       if (opt->md.dangles == 1)
         opt->md.dangles = 2;
 
@@ -941,9 +1139,10 @@ process_record(struct record_data *record)
       if (opt->md.compute_bpp) {
         vrna_pf_dimer_probs(AB.F0AB, AB.FA, AB.FB, prAB, prA, prB, Alength, vc->exp_params);
         vrna_pf_dimer_probs(AA.F0AB, AA.FA, AA.FA, prAA, prA, prA, Alength, vc->exp_params);
-        vrna_pf_dimer_probs(BB.F0AB, BB.FA, BB.FA, prBB, prA, prB, Blength, vc->exp_params);
+        vrna_pf_dimer_probs(BB.F0AB, BB.FA, BB.FA, prBB, prB, prB, Blength, vc->exp_params);
       }
 
+#endif
       if (opt->doC) {
         conc_result = vrna_pf_dimer_concentrations(AB.FcAB,
                                                    AA.FcAB,
@@ -953,26 +1152,7 @@ process_record(struct record_data *record)
                                                    concentrations,
                                                    vc->exp_params);
       }
-    }
 
-    if (opt->csv_output) {
-      vrna_cstr_printf(o_stream->data,
-                       "%c"
-                       "%g%c"   /* probability of MFE structure */
-                       "%6.2f", /* delta G binding */
-                       opt->csv_output_delim,
-                       exp((AB.FAB - min_en) / kT), opt->csv_output_delim,
-                       AB.FcAB - AB.FA - AB.FB);
-    } else {
-      vrna_cstr_printf_structure(o_stream->data,
-                                 NULL,
-                                 " frequency of mfe structure in ensemble %g"
-                                 "; delta G binding=%6.2f",
-                                 exp((AB.FAB - min_en) / kT),
-                                 AB.FcAB - AB.FA - AB.FB);
-    }
-
-    if (opt->doT) {
       if (opt->csv_output) {
         vrna_cstr_printf(o_stream->data,
                          "%c"
@@ -1012,7 +1192,7 @@ process_record(struct record_data *record)
 
         /*AB dot_plot*/
         fname_dot = vrna_strdup_printf("AB%s", filename_dot);
-        seq       = vrna_strdup_printf("%s%s", orig_Astring, orig_Bstring);
+        seq       = vrna_strdup_printf("%s&%s", orig_Astring, orig_Bstring);
         comment   = vrna_strdup_printf("Heterodimer AB FreeEnergy= %.9f", AB.FcAB);
         THREADSAFE_FILE_OUTPUT(
           (void)vrna_plot_dp_PS_list(seq,
@@ -1027,7 +1207,7 @@ process_record(struct record_data *record)
 
         /*AA dot_plot*/
         fname_dot = vrna_strdup_printf("AA%s", filename_dot);
-        seq       = vrna_strdup_printf("%s%s", orig_Astring, orig_Astring);
+        seq       = vrna_strdup_printf("%s&%s", orig_Astring, orig_Astring);
         comment   = vrna_strdup_printf("Homodimer AA FreeEnergy= %.9f", AA.FcAB);
         THREADSAFE_FILE_OUTPUT(
           (void)vrna_plot_dp_PS_list(seq,
@@ -1042,7 +1222,7 @@ process_record(struct record_data *record)
 
         /*BB dot_plot*/
         fname_dot = vrna_strdup_printf("BB%s", filename_dot);
-        seq       = vrna_strdup_printf("%s%s", orig_Bstring, orig_Bstring);
+        seq       = vrna_strdup_printf("%s&%s", orig_Bstring, orig_Bstring);
         comment   = vrna_strdup_printf("Homodimer BB FreeEnergy= %.9f", BB.FcAB);
         THREADSAFE_FILE_OUTPUT(
           (void)vrna_plot_dp_PS_list(seq,
@@ -1088,7 +1268,7 @@ process_record(struct record_data *record)
         if (filename_dot) {
           THREADSAFE_FILE_OUTPUT(
             (void)vrna_plot_dp_PS_list(record->sequence,
-                                       vc->cutpoint,
+                                       (vc->strands > 1) ? vc->strand_start[1] : -1,
                                        filename_dot,
                                        prAB,
                                        mfAB,
@@ -1161,77 +1341,50 @@ write_csv_header(FILE           *output,
 {
   vrna_cstr_t stream = vrna_cstr(100, output);
 
+  vrna_cstr_printf(stream,
+                   "seq_num%c"
+                   "seq_id%c"
+                   "seq%c"
+                   "mfe_struct%c"
+                   "mfe",
+                   opt->csv_output_delim,
+                   opt->csv_output_delim,
+                   opt->csv_output_delim,
+                   opt->csv_output_delim);
+
   /* compose header line for CSV output */
   if (opt->pf) {
     if (opt->md.compute_bpp) {
-      if (opt->doT) {
-        vrna_cstr_printf(stream,
-                         "seq_num%c"
-                         "seq_id%c"
-                         "seq%c"
-                         "mfe_struct%c"
-                         "mfe%c"
-                         "bpp_string%c"
-                         "ensemble_energy%c"
-                         "AB%c"
-                         "AA%c"
-                         "BB%c"
-                         "A%c"
-                         "B\n",
-                         opt->csv_output_delim,
-                         opt->csv_output_delim,
-                         opt->csv_output_delim,
-                         opt->csv_output_delim,
-                         opt->csv_output_delim,
-                         opt->csv_output_delim,
-                         opt->csv_output_delim,
-                         opt->csv_output_delim,
-                         opt->csv_output_delim,
-                         opt->csv_output_delim,
-                         opt->csv_output_delim);
-      } else {
-        vrna_cstr_printf(stream,
-                         "seq_num%c"
-                         "seq_id%c"
-                         "seq%c"
-                         "mfe_struct%c"
-                         "mfe%c",
-                         "bpp_string%c"
-                         "ensemble_energy\n",
-                         opt->csv_output_delim,
-                         opt->csv_output_delim,
-                         opt->csv_output_delim,
-                         opt->csv_output_delim,
-                         opt->csv_output_delim,
-                         opt->csv_output_delim);
-      }
-    } else {
       vrna_cstr_printf(stream,
-                       "seq_num%c"
-                       "seq_id%c"
-                       "seq%c"
-                       "mfe_struct%c"
-                       "mfe%c",
-                       "ensemble_energy\n",
+                       "%cbpp_string",
+                       opt->csv_output_delim);
+    }
+
+    vrna_cstr_printf(stream,
+                     "%censemble_energy%c"
+                     "prob_mfe%c"
+                     "dG_binding",
+                     opt->csv_output_delim,
+                     opt->csv_output_delim,
+                     opt->csv_output_delim);
+
+    if (opt->doT) {
+      vrna_cstr_printf(stream,
+                       "%cAB%c"
+                       "AA%c"
+                       "BB%c"
+                       "A%c"
+                       "B",
                        opt->csv_output_delim,
                        opt->csv_output_delim,
                        opt->csv_output_delim,
                        opt->csv_output_delim,
                        opt->csv_output_delim);
     }
-  } else {
-    vrna_cstr_printf(stream,
-                     "seq_num%c"
-                     "seq_id%c"
-                     "seq%c"
-                     "mfe_struct%c"
-                     "mfe\n",
-                     opt->csv_output_delim,
-                     opt->csv_output_delim,
-                     opt->csv_output_delim,
-                     opt->csv_output_delim);
   }
 
+  vrna_cstr_printf(stream, "\n");
+
   vrna_cstr_fflush(stream);
   vrna_cstr_close(stream);
 }
@@ -1242,7 +1395,7 @@ compute_MEA(vrna_fold_compound_t  *fc,
             double                MEAgamma,
             vrna_cstr_t           rec_output)
 {
-  char  *structure, *mea_structure;
+  char  *structure, *mea_structure = NULL;
   float mea, mea_en;
 
   structure = vrna_MEA(fc, MEAgamma, &mea);
@@ -1250,7 +1403,17 @@ compute_MEA(vrna_fold_compound_t  *fc,
   mea_en = vrna_eval_structure(fc, (const char *)structure);
 
   /* insert cut point */
-  mea_structure = vrna_cut_point_insert(structure, fc->cutpoint);
+  char  *tmp_struct = strdup(structure);
+
+  if (fc->strands == 1) {
+    mea_structure = tmp_struct;
+  } else {
+    for (unsigned int i = 1; i < fc->strands; i++) {
+      mea_structure = vrna_cut_point_insert(tmp_struct, (int)fc->strand_start[i] + (i - 1));
+      free(tmp_struct);
+      tmp_struct = mea_structure;
+    }
+  }
 
   vrna_cstr_printf_structure(rec_output, mea_structure, " {%6.2f MEA=%.2f}", mea_en, mea);
 
@@ -1263,14 +1426,24 @@ static void
 compute_centroid(vrna_fold_compound_t *fc,
                  vrna_cstr_t          rec_output)
 {
-  char    *cent, *centroid_structure;
+  char    *cent, *centroid_structure = NULL;
   double  cent_en, dist;
 
   cent    = vrna_centroid(fc, &dist);
   cent_en = vrna_eval_structure(fc, (const char *)cent);
 
   /* insert cut point */
-  centroid_structure = vrna_cut_point_insert(cent, fc->cutpoint);
+  char  *tmp_struct = strdup(cent);
+
+  if (fc->strands == 1) {
+    centroid_structure = tmp_struct;
+  } else {
+    for (unsigned int i = 1; i < fc->strands; i++) {
+      centroid_structure = vrna_cut_point_insert(tmp_struct, (int)fc->strand_start[i] + (i - 1));
+      free(tmp_struct);
+      tmp_struct = centroid_structure;
+    }
+  }
 
   vrna_cstr_printf_structure(rec_output, centroid_structure, " {%6.2f d=%.2f}", cent_en, dist);
 
@@ -1397,11 +1570,11 @@ postscript_layout(vrna_fold_compound_t  *fc,
     filename_plot = strdup("rna.ps");
   }
 
-  if (fc->cutpoint >= 0) {
+  if (fc->strands > 1) {
     annot = vrna_strdup_printf("1 %d 9  0 0.9 0.2 omark\n"
                                "%d %d 9  1 0.1 0.2 omark\n",
-                               fc->cutpoint - 1,
-                               fc->cutpoint + 1,
+                               fc->strand_end[0],
+                               fc->strand_start[1] + 1,
                                fc->length + 1);
   }
 
