@@ -15,44 +15,91 @@
 #include <ctype.h>
 #include <unistd.h>
 #include <string.h>
-#include "ViennaRNA/fold.h"
-#include "ViennaRNA/part_func.h"
+#include "ViennaRNA/mfe/local.h"
+#include "ViennaRNA/partfunc/local.h"
 #include "ViennaRNA/fold_vars.h"
-#include "ViennaRNA/utils.h"
-#include "ViennaRNA/PS_dot.h"
-#include "ViennaRNA/energy_const.h"
-#include "ViennaRNA/read_epars.h"
+#include "ViennaRNA/utils/basic.h"
+#include "ViennaRNA/utils/strings.h"
+#include "ViennaRNA/utils/log.h"
+#include "ViennaRNA/plotting/probabilities.h"
+#include "ViennaRNA/params/constants.h"
+#include "ViennaRNA/params/io.h"
 #include "ViennaRNA/LPfold.h"
-#include "ViennaRNA/params.h"
-#include "ViennaRNA/file_formats.h"
-#include "ViennaRNA/commands.h"
-#include "RNAplfold_cmdl.h"
-#include "gengetopt_helper.h"
-#include "input_id_helpers.h"
+#include "ViennaRNA/params/basic.h"
+#include "ViennaRNA/probing/SHAPE.h"
+#include "ViennaRNA/constraints/soft_special.h"
+#include "ViennaRNA/io/file_formats.h"
+#include "ViennaRNA/io/utils.h"
+#include "ViennaRNA/io/commands.h"
 
-#include "ViennaRNA/color_output.inc"
+#include "RNAplfold_cmdl.h"
+#include "gengetopt_helpers.h"
+#include "input_id_helpers.h"
+#include "modified_bases_helpers.h"
+#include "probing_data_helpers.h"
+
+#include "ViennaRNA/intern/color_output.h"
+
+#ifndef isnan
+#define isnan(x) \
+  (sizeof(x) == sizeof(long double) ? isnan_ld(x) \
+   : sizeof(x) == sizeof(double) ? isnan_d(x) \
+   : isnan_f(x))
+
+/* Use volatile tmp variable to prevent optimization by compiler. */
+static inline int
+isnan_f(float x)
+{
+  volatile float tmp = x;
+
+  return tmp != x;
+}
+
+
+static inline int
+isnan_d(double x)
+{
+  volatile double tmp = x;
+
+  return tmp != x;
+}
+
+
+static inline int
+isnan_ld(long double x)
+{
+  volatile long double tmp = x;
+
+  return tmp != x;
+}
+
+
+#endif /* ifndef isnan */
 
 typedef struct {
-  float cutoff;
-  FILE  *pUfp;
-  FILE  *spup;
+  float     cutoff;
+  FILE      *pUfp;
+  FILE      *spup;
   vrna_ep_t *plist;
-  int        plist_cnt;
-  int   plexoutput;
-  int   simply_putout;
-  int   openenergies;
-  double  **pup;
-  int     ulength;
-  int     n;
-  double  kT;
+  int       plist_cnt;
+  int       plexoutput;
+  int       simply_putout;
+  int       openenergies;
+  double    **pup;
+  int       ulength;
+  int       n;
+  double    kT;
 } plfold_data;
 
 int unpaired;
-PRIVATE void putoutphakim_u(vrna_fold_compound_t *fc,
-                            double  **pU,
-                            int     length,
-                            int     ulength,
-                            FILE    *fp);
+
+PRIVATE void
+putoutphakim_u(vrna_fold_compound_t *fc,
+               double               **pU,
+               int                  length,
+               int                  ulength,
+               FILE                 *fp);
+
 
 PRIVATE void
 plfold_callback(FLT_OR_DBL    *pr,
@@ -62,21 +109,32 @@ plfold_callback(FLT_OR_DBL    *pr,
                 unsigned int  type,
                 void          *data);
 
+
 PRIVATE void
 prepare_up_file(plfold_data *data);
 
+
 PRIVATE void
-print_up_open(FILE *fp, int i, double *pr, int pr_size, int ulength, double kT);
+print_up_open(FILE    *fp,
+              int     i,
+              double  *pr,
+              int     pr_size,
+              int     ulength,
+              double  kT);
 
 
 PRIVATE void
-print_up(FILE *fp, int i, double *pr, int pr_size, int ulength);
+print_up(FILE   *fp,
+         int    i,
+         double *pr,
+         int    pr_size,
+         int    ulength);
 
 
 PRIVATE void
 print_pu_bin(vrna_fold_compound_t *fc,
-             plfold_data *data,
-             int ulength);
+             plfold_data          *data,
+             int                  ulength);
 
 
 /*--------------------------------------------------------------------------*/
@@ -84,45 +142,41 @@ int
 main(int  argc,
      char *argv[])
 {
-  FILE                        *pUfp, *spup;
+  FILE                        *pUfp;
   struct RNAplfold_args_info  args_info;
   char                        *structure, *ParamFile, *ns_bases, *rec_sequence, *rec_id,
-                              **rec_rest, *orig_sequence, *id_prefix, *id_delim, *filename_delim,
-                              *command_file;
+                              **rec_rest, *orig_sequence, *filename_delim, *command_file,
+                              *shape_file, *shape_method, *shape_conversion;
   unsigned int                rec_type, read_opt;
   int                         length, istty, winsize, pairdist, tempwin, temppair, tempunpaired,
-                              noconv, i, plexoutput, simply_putout, openenergies, binaries, auto_id,
-                              id_digits, filename_full;
-  long int                    seq_number;
+                              noconv, i, plexoutput, simply_putout, openenergies, binaries,
+                              filename_full, with_shapes, verbose;
   float                       cutoff;
-  double                      **pup, betaScale;
-  vrna_ep_t                   *pl, *dpp;
   vrna_exp_param_t            *pf_parameters;
   vrna_md_t                   md;
-  vrna_cmd_t                  *commands;
+  vrna_cmd_t                  commands;
+  dataset_id                  id_control;
+  vrna_sc_mod_param_t         *mod_params;
 
-  pUfp          = NULL;
-  spup          = NULL;
-  pup           = NULL; /*prob of being unpaired, lengthwise*/
-  dpp           = NULL;
-  dangles       = 2;
-  cutoff        = 0.01;
-  winsize       = 70;
-  pairdist      = 0;
-  unpaired      = 0;
-  betaScale     = 1.;
-  simply_putout = plexoutput = openenergies = noconv = 0;
-  binaries      = 0;
-  tempwin       = temppair = tempunpaired = 0;
-  structure     = ParamFile = ns_bases = NULL;
-  rec_type      = read_opt = 0;
-  rec_id        = rec_sequence = orig_sequence = NULL;
-  rec_rest      = NULL;
-  pf_parameters = NULL;
-  auto_id       = 0;
-  filename_full = 0;
-  command_file  = NULL;
-  commands      = NULL;
+  pUfp                = NULL;
+  dangles             = 2;
+  cutoff              = 0.01;
+  winsize             = 70;
+  pairdist            = 0;
+  unpaired            = 0;
+  simply_putout       = plexoutput = openenergies = noconv = 0;
+  binaries            = 0;
+  tempwin             = temppair = tempunpaired = 0;
+  structure           = ParamFile = ns_bases = NULL;
+  rec_type            = read_opt = 0;
+  rec_id              = rec_sequence = orig_sequence = NULL;
+  rec_rest            = NULL;
+  pf_parameters       = NULL;
+  filename_full       = 0;
+  command_file        = NULL;
+  commands            = NULL;
+  verbose             = 0;
+  mod_params          = NULL;
 
   set_model_details(&md);
 
@@ -134,13 +188,16 @@ main(int  argc,
   if (RNAplfold_cmdline_parser(argc, argv, &args_info) != 0)
     exit(1);
 
+  /* prepare logging system and verbose mode */
+  ggo_log_settings(args_info, verbose);
+
+  /* SHAPE reactivity data */
+  ggo_get_SHAPE(args_info, with_shapes, shape_file, shape_method, shape_conversion);
+
   /* parse options for ID manipulation */
-  ggo_get_ID_manipulation(args_info,
-                          auto_id,
-                          id_prefix, "sequence",
-                          id_delim, "_",
-                          id_digits, 4,
-                          seq_number, 1);
+  ggo_get_id_control(args_info, id_control, "Sequence", "sequence", "_", 4, 1);
+
+  ggo_get_md_part(args_info, md);
 
   /* temperature */
   if (args_info.temp_given)
@@ -153,7 +210,7 @@ main(int  argc,
   /* set dangle model */
   if (args_info.dangles_given) {
     if ((args_info.dangles_arg != 0) && (args_info.dangles_arg != 2))
-      vrna_message_warning(
+      vrna_log_warning(
         "required dangle model not implemented, falling back to default dangles=2");
     else
       md.dangles = dangles = args_info.dangles_arg;
@@ -167,6 +224,9 @@ main(int  argc,
   if (args_info.noGU_given)
     md.noGU = noGU = 1;
 
+  if (args_info.salt_given)
+    md.salt = args_info.salt_arg;
+
   /* do not allow weak closing pairs (AU,GU) */
   if (args_info.noClosingGU_given)
     md.noGUclosure = no_closingGU = 1;
@@ -180,8 +240,7 @@ main(int  argc,
     md.energy_set = energy_set = args_info.energyModel_arg;
 
   /* take another energy parameter set */
-  if (args_info.paramFile_given)
-    ParamFile = strdup(args_info.paramFile_arg);
+  ggo_get_read_paramFile(args_info, &md);
 
   /* Allow other pairs in addition to the usual AU,GC,and GU pairs */
   if (args_info.nsp_given)
@@ -219,9 +278,6 @@ main(int  argc,
   if (args_info.binaries_given)
     binaries = 1;
 
-  if (args_info.betaScale_given)
-    md.betaScale = betaScale = args_info.betaScale_arg;
-
   /* check for errorneous parameter options */
   if ((pairdist < 0) || (cutoff < 0.) || (unpaired < 0) || (winsize < 0)) {
     RNAplfold_cmdline_parser_print_help();
@@ -231,10 +287,12 @@ main(int  argc,
   /* filename sanitize delimiter */
   if (args_info.filename_delim_given)
     filename_delim = strdup(args_info.filename_delim_arg);
+  else if (get_id_delim(id_control))
+    filename_delim = strdup(get_id_delim(id_control));
   else
-    filename_delim = strdup(id_delim);
+    filename_delim = NULL;
 
-  if (isspace(*filename_delim)) {
+  if ((filename_delim) && isspace(*filename_delim)) {
     free(filename_delim);
     filename_delim = NULL;
   }
@@ -246,6 +304,12 @@ main(int  argc,
   if (args_info.commands_given)
     command_file = strdup(args_info.commands_arg);
 
+  ggo_get_modified_base_settings(args_info,
+                                 mod_params,
+                                 &(md));
+
+  ggo_geometry_settings(args_info, &md);
+
   /* free allocated memory of command line data structure */
   RNAplfold_cmdline_parser_free(&args_info);
 
@@ -254,9 +318,6 @@ main(int  argc,
    # begin initializing
    #############################################
    */
-  if (ParamFile != NULL)
-    read_parameter_file(ParamFile);
-
   if (ns_bases != NULL)
     vrna_md_set_nonstandards(&md, ns_bases);
 
@@ -271,14 +332,14 @@ main(int  argc,
     pairdist = winsize;
 
   if (pairdist > winsize) {
-    vrna_message_warning("pairdist (-L %d) should be <= winsize (-W %d);"
+    vrna_log_warning("pairdist (-L %d) should be <= winsize (-W %d);"
                          "Setting pairdist=winsize",
                          pairdist, winsize);
     pairdist = winsize;
   }
 
   if (dangles % 2) {
-    vrna_message_warning("using default dangles = 2");
+    vrna_log_warning("using default dangles = 2");
     md.dangles = dangles = 2;
   }
 
@@ -297,7 +358,9 @@ main(int  argc,
   while (
     !((rec_type = vrna_file_fasta_read_record(&rec_id, &rec_sequence, &rec_rest, NULL, read_opt))
       & (VRNA_INPUT_ERROR | VRNA_INPUT_QUIT))) {
-    char *SEQ_ID = NULL;
+    size_t  **mod_positions, mod_param_sets;
+    char    *SEQ_ID = NULL;
+
     /*
      ########################################################
      # init everything according to the data we've read
@@ -307,11 +370,18 @@ main(int  argc,
       rec_id = memmove(rec_id, rec_id + 1, strlen(rec_id));
 
     /* construct the sequence ID */
-    ID_generate(SEQ_ID, rec_id, auto_id, id_prefix, id_delim, id_digits, seq_number, filename_full);
+    set_next_id(&rec_id, id_control);
+    SEQ_ID = fileprefix_from_id(rec_id, id_control, filename_full);
 
     length    = (int)strlen(rec_sequence);
     structure = (char *)vrna_alloc(sizeof(char) * (length + 1));
 
+
+    mod_positions = mod_positions_seq_prepare(rec_sequence,
+                                              mod_params,
+                                              verbose,
+                                              &mod_param_sets);
+
     /* convert DNA alphabet to RNA if not explicitely switched off */
     if (!noconv)
       vrna_seq_toRNA(rec_sequence);
@@ -322,7 +392,7 @@ main(int  argc,
     vrna_seq_toupper(rec_sequence);
 
     if (istty)
-      vrna_message_info(stdout, "length = %d", length);
+      printf("length = %d", length);
 
     /*
      ########################################################
@@ -332,19 +402,19 @@ main(int  argc,
 
     if (length > 1000000) {
       if (!simply_putout && !unpaired) {
-        vrna_message_warning("Switched to simple output mode!!!");
+        vrna_log_warning("Switched to simple output mode!!!");
         simply_putout = 1;
       }
     }
 
     if ((simply_putout) && (plexoutput)) {
-      vrna_message_warning("plexoutput not available in simple output mode!\n"
+      vrna_log_warning("plexoutput not available in simple output mode!\n"
                            "Switching back to full mode instead!");
       simply_putout = 0;
     }
 
     if ((simply_putout) && (binaries)) {
-      vrna_message_warning("binary output not available in simple output mode!\n"
+      vrna_log_warning("binary output not available in simple output mode!\n"
                            "Switching back to full mode instead!");
       simply_putout = 0;
     }
@@ -369,7 +439,7 @@ main(int  argc,
 
     /* adjust winsize, pairdist and ulength if necessary */
     if (length < winsize) {
-      vrna_message_warning("window size %d larger than sequence length %d", winsize, length);
+      vrna_log_warning("window size %d larger than sequence length %d", winsize, length);
       tempwin = winsize;
       winsize = length;
       if (pairdist > winsize) {
@@ -396,15 +466,18 @@ main(int  argc,
       if (!SEQ_ID)
         SEQ_ID = strdup("plfold");
 
-      fname1  = vrna_strdup_printf("%s%slunp", SEQ_ID, id_delim);
-      fname2  = vrna_strdup_printf("%s%sbasepairs", SEQ_ID, id_delim);
-      fname3  = vrna_strdup_printf("%s%suplex", SEQ_ID, id_delim);
-      fname4  =
-        (binaries) ? vrna_strdup_printf("%s%sopenen%sbin", SEQ_ID, id_delim,
-                                        id_delim) : vrna_strdup_printf("%s%sopenen",
-                                                                       SEQ_ID,
-                                                                       id_delim);
-      ffname = vrna_strdup_printf("%s%sdp.ps", SEQ_ID, id_delim);
+      fname1  = vrna_strdup_printf("%s%slunp", SEQ_ID, filename_delim);
+      fname2  = vrna_strdup_printf("%s%sbasepairs", SEQ_ID, filename_delim);
+      fname3  = vrna_strdup_printf("%s%suplex", SEQ_ID, filename_delim);
+      fname4  = (binaries) ?
+                vrna_strdup_printf("%s%sopenen%sbin",
+                                   SEQ_ID,
+                                   filename_delim,
+                                   filename_delim) :
+                vrna_strdup_printf("%s%sopenen",
+                                   SEQ_ID,
+                                   filename_delim);
+      ffname = vrna_strdup_printf("%s%sdp.ps", SEQ_ID, filename_delim);
 
       /* sanitize filenames */
       tmp_string = vrna_filename_sanitize(fname1, filename_delim);
@@ -424,21 +497,36 @@ main(int  argc,
       ffname = tmp_string;
 
 
-      md.compute_bpp = 1;
-      md.window_size = winsize;
-      md.max_bp_span = pairdist;
+      md.compute_bpp  = 1;
+      md.window_size  = winsize;
+      md.max_bp_span  = pairdist;
 
       vrna_fold_compound_t *fc = vrna_fold_compound(rec_sequence, &md, VRNA_OPTION_WINDOW);
 
+      if (with_shapes) {
+        vrna_constraints_add_SHAPE(fc,
+                                   shape_file,
+                                   shape_method,
+                                   shape_conversion,
+                                   verbose,
+                                   VRNA_OPTION_DEFAULT | VRNA_OPTION_WINDOW);
+      }
+
       if (commands)
         vrna_commands_apply(fc, commands, VRNA_CMD_PARSE_HC | VRNA_CMD_PARSE_SC);
 
+      /* apply modified base support if requested */
+      mod_bases_apply(fc,
+                      mod_param_sets,
+                      mod_positions,
+                      mod_params);
+
       pf_parameters = vrna_exp_params(&md);
 
       /* prepare data structure for callback */
       plfold_data data;
 
-      data.cutoff = cutoff;
+      data.cutoff         = cutoff;
       data.spup           = (simply_putout) ? fopen(fname2, "w") : NULL;
       data.plexoutput     = plexoutput;
       data.simply_putout  = simply_putout;
@@ -462,8 +550,8 @@ main(int  argc,
           data.pUfp       = NULL;
         }
       } else {
-        data.pup        = NULL;
-        data.pUfp       = NULL;
+        data.pup  = NULL;
+        data.pUfp = NULL;
       }
 
       /* prepare option flags */
@@ -476,7 +564,13 @@ main(int  argc,
         plfold_opt |= VRNA_PROBS_WINDOW_UP;
 
       /* perform recursions */
-      vrna_probs_window(fc, unpaired, plfold_opt, &plfold_callback, (void *)&data);
+      int r = vrna_probs_window(fc, unpaired, plfold_opt, &plfold_callback, (void *)&data);
+
+      if (!r) {
+        vrna_log_warning("Something bad happened while processing the input! "
+                             "Aborting now...");
+        goto rnaplfold_exit;
+      }
 
       if (!simply_putout) {
         /* create dot plot output */
@@ -492,18 +586,22 @@ main(int  argc,
 
           /* print unpaired probabilities to file */
 
+          data.pUfp = fopen(openenergies ? fname4 : fname1, "w");
           if (binaries) {
-            data.pUfp = fopen(fname1, "w");
             print_pu_bin(fc, &data, unpaired);
           } else {
-            data.pUfp = fopen(openenergies ? fname4 : fname1, "w");
             prepare_up_file(&data);
             if (openenergies) {
               for (i = 1; i <= length; i++)
-                 print_up_open(data.pUfp, i, data.pup[i], (i > unpaired) ? unpaired : i, unpaired, data.kT / 1000.);
+                print_up_open(data.pUfp,
+                              i,
+                              data.pup[i],
+                              (i > unpaired) ? unpaired : i,
+                              unpaired,
+                              data.kT / 1000.);
             } else {
               for (i = 1; i <= length; i++)
-                 print_up(data.pUfp, i, data.pup[i], (i > unpaired) ? unpaired : i, unpaired);
+                print_up(data.pUfp, i, data.pup[i], (i > unpaired) ? unpaired : i, unpaired);
             }
           }
 
@@ -546,34 +644,49 @@ main(int  argc,
     free(structure);
     free(SEQ_ID);
 
+    if (with_shapes)
+      break;
+
     rec_id    = rec_sequence = orig_sequence = NULL;
     rec_rest  = NULL;
 
-    ID_number_increase(&seq_number, "Sequence");
-
     /* print user help for the next round if we get input from tty */
     if (istty)
       vrna_message_input_seq_simple();
   }
 
-  free(id_prefix);
-  free(id_delim);
+rnaplfold_exit:
+
   free(filename_delim);
   free(command_file);
+  free(shape_method);
+  free(shape_conversion);
   vrna_commands_free(commands);
 
+  if (mod_params) {
+    for (vrna_sc_mod_param_t *ptr = mod_params; *ptr != NULL; ptr++)
+      vrna_sc_mod_parameters_free(*ptr);
+
+    free(mod_params);
+  }
+
+  free_id_data(id_control);
+
+  if (vrna_log_fp() != stderr)
+    fclose(vrna_log_fp());
+
   return EXIT_SUCCESS;
 }
 
 
 PRIVATE void
 print_pu_bin(vrna_fold_compound_t *fc,
-             plfold_data *data,
-             int ulength)
+             plfold_data          *data,
+             int                  ulength)
 {
-  unsigned int length;
-  int i, k, *p;
-  double  kT = fc->exp_params->kT / 1000.0;
+  unsigned int  length;
+  int           i, k, *p;
+  double        kT = fc->exp_params->kT / 1000.0;
 
   length = fc->length;
 
@@ -617,18 +730,20 @@ print_pu_bin(vrna_fold_compound_t *fc,
   free(p);
 }
 
+
 PRIVATE void
 prepare_up_file(plfold_data *data)
 {
-  int i, *p;
+  int i;
 
-  if (data->openenergies) {
+  if (data->openenergies)
     fprintf(data->pUfp, "#opening energies\n #i$\tl=");
-  } else {
+  else
     fprintf(data->pUfp, "#unpaired probabilities\n #i$\tl=");
-  }
+
   for (i = 1; i <= data->ulength; i++)
     fprintf(data->pUfp, "%d\t", i);
+
   fprintf(data->pUfp, "\n");
 }
 
@@ -641,7 +756,7 @@ plfold_callback(FLT_OR_DBL    *pr,
                 unsigned int  type,
                 void          *data)
 {
-  int cnt;
+  int         cnt;
   plfold_data *d;
 
   d = (plfold_data *)data;
@@ -649,13 +764,14 @@ plfold_callback(FLT_OR_DBL    *pr,
   if (type & VRNA_PROBS_WINDOW_BPP) {
     if (!d->simply_putout) {
       /* store pair probabilities in plist */
-      d->plist = (vrna_ep_t *)vrna_realloc(d->plist, sizeof(vrna_ep_t) * (d->plist_cnt + pr_size + 1));
+      d->plist = (vrna_ep_t *)vrna_realloc(d->plist,
+                                           sizeof(vrna_ep_t) * (d->plist_cnt + pr_size + 1));
 
       for (cnt = i + 1; cnt <= pr_size; cnt++) {
         if (pr[cnt] >= d->cutoff) {
-          d->plist[d->plist_cnt].i = i;
-          d->plist[d->plist_cnt].j = cnt;
-          d->plist[d->plist_cnt].p = pr[cnt];
+          d->plist[d->plist_cnt].i    = i;
+          d->plist[d->plist_cnt].j    = cnt;
+          d->plist[d->plist_cnt].p    = pr[cnt];
           d->plist[d->plist_cnt].type = VRNA_PLIST_TYPE_BASEPAIR;
           (d->plist_cnt)++;
         }
@@ -665,28 +781,26 @@ plfold_callback(FLT_OR_DBL    *pr,
       d->plist = (vrna_ep_t *)vrna_realloc(d->plist, sizeof(vrna_ep_t) * (d->plist_cnt + 1));
 
       /* add end-marker to last element */
-      d->plist[d->plist_cnt].i = 0;
-      d->plist[d->plist_cnt].j = 0;
-      d->plist[d->plist_cnt].p = 0.;
+      d->plist[d->plist_cnt].i    = 0;
+      d->plist[d->plist_cnt].j    = 0;
+      d->plist[d->plist_cnt].p    = 0.;
       d->plist[d->plist_cnt].type = VRNA_PLIST_TYPE_BASEPAIR;
     } else {
       /* print pair probabilities to output file handle */
-      for (cnt = i + 1; cnt <= pr_size; cnt++) {
-        if (pr[cnt] >= d->cutoff) {
+      for (cnt = i + 1; cnt <= pr_size; cnt++)
+        if (pr[cnt] >= d->cutoff)
           fprintf(d->spup, "%d  %d  %g\n", i, cnt, pr[cnt]);
-        }
-      }
     }
   }
 
   /* limit output to full unpaired probabilities */
-  if ((type & VRNA_PROBS_WINDOW_UP)  && ((type & VRNA_ANY_LOOP) == VRNA_ANY_LOOP)) {
+  if ((type & VRNA_PROBS_WINDOW_UP) && ((type & VRNA_ANY_LOOP) == VRNA_ANY_LOOP)) {
     if (!d->simply_putout) {
       /* store unpaired probabilities in an array */
 
       /* first allocate some memory */
-      d->pup[i] = (double *)vrna_realloc(d->pup[i], sizeof(double) * (max + 1));
-      d->pup[i][0] = 0.;
+      d->pup[i]     = (double *)vrna_realloc(d->pup[i], sizeof(double) * (max + 1));
+      d->pup[i][0]  = 0.;
       /* copy over unpaired probabilities */
       for (cnt = 1; cnt <= pr_size; cnt++)
         d->pup[i][cnt] = pr[cnt];
@@ -704,14 +818,28 @@ plfold_callback(FLT_OR_DBL    *pr,
 
 
 PRIVATE void
-print_up_open(FILE *fp, int i, double *pr, int pr_size, int ulength, double kT)
+print_up_open(FILE    *fp,
+              int     i,
+              double  *pr,
+              int     pr_size,
+              int     ulength,
+              double  kT)
 {
   int cnt;
 
   fprintf(fp, "%d\t", i);
-  for (cnt = 1; cnt < pr_size; cnt++)
-    fprintf(fp, "%.7g\t", -log(pr[cnt]) * kT);
-  fprintf(fp, "%.7g", -log(pr[pr_size]) * kT);
+  for (cnt = 1; cnt < pr_size; cnt++) {
+    if (isnan(pr[cnt]) || (pr[cnt] == 0.))
+      fprintf(fp, "NA\t");
+    else
+      fprintf(fp, "%.7g\t", -log(pr[cnt]) * kT);
+  }
+
+  if (isnan(pr[pr_size]) || (pr[pr_size] == 0.))
+    fprintf(fp, "NA");
+  else
+    fprintf(fp, "%.7g", -log(pr[pr_size]) * kT);
+
   for (cnt = pr_size + 1; cnt <= ulength; cnt++)
     fprintf(fp, "\tNA");
   fprintf(fp, "\n");
@@ -719,14 +847,27 @@ print_up_open(FILE *fp, int i, double *pr, int pr_size, int ulength, double kT)
 
 
 PRIVATE void
-print_up(FILE *fp, int i, double *pr, int pr_size, int ulength)
+print_up(FILE   *fp,
+         int    i,
+         double *pr,
+         int    pr_size,
+         int    ulength)
 {
   int cnt;
 
   fprintf(fp, "%d\t", i);
-  for (cnt = 1; cnt < pr_size; cnt++)
-    fprintf(fp, "%.7g\t", pr[cnt]);
-  fprintf(fp, "%.7g", pr[pr_size]);
+  for (cnt = 1; cnt < pr_size; cnt++) {
+    if (isnan(pr[cnt]))
+      fprintf(fp, "NA\t");
+    else
+      fprintf(fp, "%.7g\t", pr[cnt]);
+  }
+
+  if (isnan(pr[pr_size]))
+    fprintf(fp, "NA");
+  else
+    fprintf(fp, "%.7g", pr[pr_size]);
+
   for (cnt = pr_size + 1; cnt <= ulength; cnt++)
     fprintf(fp, "\tNA");
   fprintf(fp, "\n");
@@ -735,10 +876,10 @@ print_up(FILE *fp, int i, double *pr, int pr_size, int ulength)
 
 PRIVATE void
 putoutphakim_u(vrna_fold_compound_t *fc,
-               double **pU,
-               int    length,
-               int    ulength,
-               FILE   *fp)
+               double               **pU,
+               int                  length,
+               int                  ulength,
+               FILE                 *fp)
 {
   /*put out Fopen in dekacalories per mol, and F(cond,open) also in dekacal*/
   int   k;
@@ -754,13 +895,15 @@ putoutphakim_u(vrna_fold_compound_t *fc,
   for (k = 1; k <= length; k++) {
     fprintf(fp, "%d\t", k);
     p0  = pU[k][1];
-    f0  = (int)-RT *log(p0) / 10;
+    f0  = (int)-RT *
+          log(p0) / 10;
 
 
     fprintf(fp, "%d\t", f0);
     if (k > 1) {
       pdep  = pU[k][2] / pU[k - 1][1];
-      fdep  = (int)-RT *log(pdep) / 10;
+      fdep  = (int)-RT *
+              log(pdep) / 10;
 
 
       fprintf(fp, "%d\t", fdep);
@@ -770,7 +913,8 @@ putoutphakim_u(vrna_fold_compound_t *fc,
 
     if (k < length) {
       pdep  = pU[k + 1][2] / pU[k + 1][1];
-      fdep  = (int)-RT *log(pdep) / 10;
+      fdep  = (int)-RT *
+              log(pdep) / 10;
 
 
       fprintf(fp, "%d\t", fdep);
