@@ -17,28 +17,37 @@
 #include <math.h>
 #include <ctype.h>
 #include <string.h>
-#include "ViennaRNA/part_func.h"
-#include "ViennaRNA/fold.h"
-#include "ViennaRNA/cofold.h"
+#include "ViennaRNA/partfunc/global.h"
+#include "ViennaRNA/sampling/basic.h"
+#include "ViennaRNA/mfe/global.h"
+#include "ViennaRNA/eval/structures.h"
 #include "ViennaRNA/fold_vars.h"
 #include "ViennaRNA/utils/basic.h"
 #include "ViennaRNA/utils/strings.h"
+#include "ViennaRNA/utils/log.h"
 #include "ViennaRNA/params/io.h"
-#include "ViennaRNA/subopt.h"
+#include "ViennaRNA/subopt/wuchty.h"
+#include "ViennaRNA/subopt/zuker.h"
 #include "ViennaRNA/params/basic.h"
 #include "ViennaRNA/constraints/basic.h"
-#include "ViennaRNA/constraints/SHAPE.h"
+#include "ViennaRNA/probing/SHAPE.h"
+#include "ViennaRNA/constraints/soft_special.h"
 #include "ViennaRNA/io/file_formats.h"
 #include "ViennaRNA/io/utils.h"
-#include "ViennaRNA/commands.h"
+#include "ViennaRNA/io/commands.h"
+#include "ViennaRNA/datastructures/string.h"
+
 #include "RNAsubopt_cmdl.h"
-#include "gengetopt_helper.h"
+#include "gengetopt_helpers.h"
 #include "input_id_helpers.h"
+#include "modified_bases_helpers.h"
+#include "probing_data_helpers.h"
 
-#include "ViennaRNA/color_output.inc"
+#include "ViennaRNA/intern/color_output.h"
 
-PRIVATE void putoutzuker(FILE                   *output,
-                         vrna_subopt_solution_t *zukersolution);
+PRIVATE void
+putoutzuker(FILE                    *output,
+            vrna_subopt_solution_t  *zukersolution);
 
 
 struct nr_en_data {
@@ -67,34 +76,36 @@ main(int  argc,
   struct          RNAsubopt_args_info args_info;
   char                                *rec_sequence, *rec_id,
                                       **rec_rest, *orig_sequence, *constraints_file, *cstruc,
-                                      *structure, *shape_file, *shape_method, *shape_conversion,
-                                      *infile, *outfile, *filename_delim;
+                                      *structure, *infile, *outfile, *filename_delim;
   unsigned int                        rec_type, read_opt;
   int                                 i, length, cl, istty, delta, n_back, noconv, dos, zuker,
-                                      with_shapes, verbose, enforceConstraints, st_back_en, batch,
+                                      verbose, enforceConstraints, st_back_en, batch,
                                       tofile, filename_full, canonicalBPonly, nonRedundant;
   double                              deltap;
   vrna_md_t                           md;
   dataset_id                          id_control;
   vrna_cmd_t                          commands;
+  vrna_sc_mod_param_t                 *mod_params;
+  probing_data_t                      *probing_data;
 
-  do_backtrack    = 1;
-  delta           = 100;
-  deltap          = n_back = noconv = dos = zuker = 0;
-  rec_type        = read_opt = 0;
-  rec_id          = rec_sequence = orig_sequence = NULL;
-  rec_rest        = NULL;
-  cstruc          = structure = NULL;
-  verbose         = 0;
-  st_back_en      = 0;
-  infile          = NULL;
-  outfile         = NULL;
-  output          = NULL;
-  tofile          = 0;
-  filename_full   = 0;
-  canonicalBPonly = 0;
-  commands        = NULL;
-  nonRedundant    = 0;
+  do_backtrack        = 1;
+  delta               = 100;
+  deltap              = n_back = noconv = dos = zuker = 0;
+  rec_type            = read_opt = 0;
+  rec_id              = rec_sequence = orig_sequence = NULL;
+  rec_rest            = NULL;
+  cstruc              = structure = NULL;
+  st_back_en          = 0;
+  infile              = NULL;
+  outfile             = NULL;
+  output              = NULL;
+  tofile              = 0;
+  filename_full       = 0;
+  canonicalBPonly     = 0;
+  commands            = NULL;
+  nonRedundant        = 0;
+  mod_params          = NULL;
+  probing_data        = NULL;
 
   set_model_details(&md);
 
@@ -109,6 +120,9 @@ main(int  argc,
   if (RNAsubopt_cmdline_parser(argc, argv, &args_info) != 0)
     exit(1);
 
+  /* prepare logging system and verbose mode */
+  ggo_log_settings(args_info, verbose);
+
   /* parse options for ID manipulation */
   ggo_get_id_control(args_info, id_control, "Sequence", "sequence", "_", 4, 1);
 
@@ -123,22 +137,16 @@ main(int  argc,
 
   /* check dangle model */
   if ((md.dangles < 0) || (md.dangles > 3)) {
-    vrna_message_warning("required dangle model not implemented, falling back to default dangles=2");
+    vrna_log_warning("required dangle model not implemented, falling back to default dangles=2");
     md.dangles = dangles = 2;
   }
 
-  /* SHAPE reactivity data */
-  ggo_get_SHAPE(args_info, with_shapes, shape_file, shape_method, shape_conversion);
-
   ggo_get_constraints_settings(args_info,
                                fold_constrained,
                                constraints_file,
                                enforceConstraints,
                                batch);
 
-  if (args_info.verbose_given)
-    verbose = 1;
-
   /* enforce canonical base pairs in any case? */
   if (args_info.canonicalBPonly_given)
     canonicalBPonly = 1;
@@ -156,21 +164,30 @@ main(int  argc,
     deltap = args_info.deltaEnergyPost_arg;
 
   /* sorted output */
-  if (args_info.sorted_given)
-    subopt_sorted = 1;
+  if (args_info.sorted_given) {
+    subopt_sorted = VRNA_SORT_BY_ENERGY_LEXICOGRAPHIC_ASC;
+    if (args_info.en_only_given)
+      subopt_sorted = VRNA_SORT_BY_ENERGY_ASC;
+  }
 
   /* stochastic backtracking */
   if (args_info.stochBT_given) {
     n_back = args_info.stochBT_arg;
-    vrna_init_rand();
     md.compute_bpp = 0;
+    if (args_info.random_seed_given)
+      vrna_init_rand_seed((unsigned int)args_info.random_seed_arg);
+    else
+      vrna_init_rand();
   }
 
   if (args_info.stochBT_en_given) {
     n_back          = args_info.stochBT_en_arg;
     st_back_en      = 1;
     md.compute_bpp  = 0;
-    vrna_init_rand();
+    if (args_info.random_seed_given)
+      vrna_init_rand_seed((unsigned int)args_info.random_seed_arg);
+    else
+      vrna_init_rand();
   }
 
   /* density of states */
@@ -189,22 +206,22 @@ main(int  argc,
 
   if (zuker) {
     if (md.circ) {
-      vrna_message_warning("Sorry, zuker subopts not yet implemented for circfold");
+      vrna_log_warning("Sorry, zuker subopts not yet implemented for circfold");
       RNAsubopt_cmdline_parser_print_help();
       exit(1);
     } else if (n_back > 0) {
-      vrna_message_warning("Can't do zuker subopts and stochastic subopts at the same time");
+      vrna_log_warning("Can't do zuker subopts and stochastic subopts at the same time");
       RNAsubopt_cmdline_parser_print_help();
       exit(1);
     } else if (md.gquad) {
-      vrna_message_warning("G-quadruplex support for Zuker subopts not implemented yet");
+      vrna_log_warning("G-quadruplex support for Zuker subopts not implemented yet");
       RNAsubopt_cmdline_parser_print_help();
       exit(1);
     }
   }
 
   if (md.gquad && (n_back > 0)) {
-    vrna_message_warning("G-quadruplex support for stochastic backtracking not implemented yet");
+    vrna_log_warning("G-quadruplex support for stochastic backtracking not implemented yet");
     RNAsubopt_cmdline_parser_print_help();
     exit(1);
   }
@@ -243,6 +260,15 @@ main(int  argc,
     commands = vrna_file_commands_read(args_info.commands_arg,
                                        VRNA_CMD_PARSE_HC | VRNA_CMD_PARSE_SC);
 
+  ggo_get_modified_base_settings(args_info,
+                                 mod_params,
+                                 &(md));
+
+  ggo_geometry_settings(args_info, &md);
+
+  /* collect probing data */
+  ggo_get_probing_data(argc, argv, args_info, probing_data);
+
   /* free allocated memory of command line data structure */
   RNAsubopt_cmdline_parser_free(&args_info);
 
@@ -254,8 +280,10 @@ main(int  argc,
 
   if (infile) {
     input = fopen((const char *)infile, "r");
-    if (!input)
-      vrna_message_error("Could not read input file");
+    if (!input) {
+      vrna_log_error("Could not read input file");
+      goto exit_fail;
+    }
   } else {
     input = stdin;
   }
@@ -265,7 +293,7 @@ main(int  argc,
   /* print user help if we get input from tty */
   if (istty) {
     if (!zuker)
-      print_comment(stdout, "Use '&' to connect 2 sequences that shall form a complex.");
+      print_comment(stdout, "Use '&' to connect sequences that shall form a complex.");
 
     if (fold_constrained) {
       vrna_message_constraint_options(
@@ -292,10 +320,11 @@ main(int  argc,
   while (
     !((rec_type = vrna_file_fasta_read_record(&rec_id, &rec_sequence, &rec_rest, input, read_opt))
       & (VRNA_INPUT_ERROR | VRNA_INPUT_QUIT))) {
-    char  *SEQ_ID         = NULL;
-    char  *v_file_name    = NULL;
-    char  *tmp_string     = NULL;
-    int   maybe_multiline = 0;
+    size_t  **mod_positions, mod_param_sets;
+    char    *SEQ_ID         = NULL;
+    char    *v_file_name    = NULL;
+    char    *tmp_string     = NULL;
+    int     maybe_multiline = 0;
 
     /*
      ########################################################
@@ -325,12 +354,15 @@ main(int  argc,
       free(v_file_name);
       v_file_name = tmp_string;
 
-      if (infile && !strcmp(infile, v_file_name))
-        vrna_message_error("Input and output file names are identical");
-
+      if (infile && !strcmp(infile, v_file_name)) {
+        vrna_log_error("Input and output file names are identical");
+        goto exit_fail;
+      }
       output = fopen((const char *)v_file_name, "a");
-      if (!output)
-        vrna_message_error("Failed to open file for writing");
+      if (!output) {
+        vrna_log_error("Failed to open file for writing");
+        goto exit_fail;
+      }
     } else {
       output = stdout;
     }
@@ -341,13 +373,16 @@ main(int  argc,
 
     /* store case-unmodified sequence */
     orig_sequence = strdup(rec_sequence);
+
+    mod_positions = mod_positions_seq_prepare(rec_sequence,
+                                              mod_params,
+                                              verbose,
+                                              &mod_param_sets);
+
     /* convert sequence to uppercase letters only */
     vrna_seq_toupper(rec_sequence);
 
-    vrna_fold_compound_t *vc =
-      vrna_fold_compound(rec_sequence, &md,
-                         VRNA_OPTION_MFE | (circ ? 0 : VRNA_OPTION_HYBRID) |
-                         ((n_back > 0) ? VRNA_OPTION_PF : 0));
+    vrna_fold_compound_t *vc = vrna_fold_compound(rec_sequence, &md, VRNA_OPTION_DEFAULT);
     length = vc->length;
 
     structure = (char *)vrna_alloc(sizeof(char) * (length + 1));
@@ -357,27 +392,44 @@ main(int  argc,
       if (constraints_file) {
         vrna_constraints_add(vc, constraints_file, VRNA_OPTION_DEFAULT);
       } else {
-        cstruc = NULL;
-        int           cp        = -1;
-        unsigned int  coptions  = (maybe_multiline) ? VRNA_OPTION_MULTILINE : 0;
-        cstruc  = vrna_extract_record_rest_structure((const char **)rec_rest, 0, coptions);
-        cstruc  = vrna_cut_point_remove(cstruc, &cp);
-        if (vc->cutpoint != cp) {
-          vrna_message_error("Sequence and Structure have different cut points.\n"
-                             "sequence: %d, structure: %d",
-                             vc->cutpoint, cp);
-        }
+        char          *constraint_ptr, **constraints;
+        unsigned int  options;
 
-        cl = (cstruc) ? (int)strlen(cstruc) : 0;
+        options  = (maybe_multiline) ? VRNA_OPTION_MULTILINE : 0;
 
-        if (cl == 0)
-          vrna_message_warning("Structure constraint is missing");
-        else if (cl < length)
-          vrna_message_warning("Structure constraint is shorter than sequence");
-        else if (cl > length)
-          vrna_message_error("Structure constraint is too long");
+        constraint_ptr = vrna_extract_record_rest_structure((const char **)rec_rest, 0, options);
+        constraints = vrna_strsplit(constraint_ptr, NULL);
+        free(constraint_ptr);
+
+        /* collect strand-wise constraints */
+        if (constraints) {
+          unsigned int  i = 0, strand_cnt = 0;
+          vrna_string_t constraint;
+          
+          constraint = vrna_string_make(NULL);
+
+          for (char **ptr = constraints; *ptr != NULL; ptr++, strand_cnt++) {
+            if (strand_cnt >= vc->strands) {
+              vrna_log_error("Structure constraint contains too many strands (expected %u, got at least %u)\n",
+                                 vc->strands,
+                                 strand_cnt + 1);
+              goto exit_fail;
+            }
+
+            unsigned int l = strlen(*ptr);
+
+            if (vc->strand_end[strand_cnt] != i + l) {
+              vrna_log_error("Length of structure constraint for strand %u differs from sequence (expected %u, got %u)\n",
+                                 strand_cnt + 1,
+                                 vc->strand_end[strand_cnt] - i,
+                                 l);
+              goto exit_fail;
+            }
+            vrna_string_append_cstring(constraint, *ptr);
+
+            i += l;
+          }
 
-        if (cstruc) {
           /* convert pseudo-dot-bracket to actual hard constraints */
           unsigned int constraint_options = VRNA_CONSTRAINT_DB_DEFAULT;
 
@@ -387,11 +439,19 @@ main(int  argc,
           if (canonicalBPonly)
             constraint_options |= VRNA_CONSTRAINT_DB_CANONICAL_BP;
 
-          vrna_constraints_add(vc, (const char *)cstruc, constraint_options);
+          vrna_constraints_add(vc, (const char *)constraint, constraint_options);
+
+          vrna_string_free(constraint);
         }
       }
     }
 
+#if 1
+    if (probing_data)
+      apply_probing_data(vc,
+                         probing_data);
+
+#else
     if (with_shapes) {
       vrna_constraints_add_SHAPE(vc,
                                  shape_file,
@@ -400,20 +460,26 @@ main(int  argc,
                                  verbose,
                                  VRNA_OPTION_MFE | ((n_back > 0) ? VRNA_OPTION_PF : 0));
     }
+#endif
 
     if (commands)
       vrna_commands_apply(vc,
                           commands,
                           VRNA_CMD_PARSE_HC | VRNA_CMD_PARSE_SC);
 
+    /* apply modified base support if requested */
+    mod_bases_apply(vc,
+                    mod_param_sets,
+                    mod_positions,
+                    mod_params);
+
     if (istty) {
-      if (cut_point == -1) {
-        vrna_message_info(stdout, "length = %d", length);
+      if (vc->strands == 1) {
+        printf("length = %d", length);
       } else {
-        vrna_message_info(stdout,
-                          "length1 = %d\nlength2 = %d",
-                          cut_point - 1,
-                          length - cut_point + 1);
+        printf("length1 = %d\nlength2 = %d",
+               cut_point - 1,
+               length - cut_point + 1);
       }
     }
 
@@ -437,16 +503,21 @@ main(int  argc,
                               VRNA_PBACKTRACK_NON_REDUNDANT :
                               VRNA_PBACKTRACK_DEFAULT;
 
-      if (vc->cutpoint != -1)
-        vrna_message_error("Boltzmann sampling for cofolded structures not implemented (yet)!");
-
+      if (vc->strands > 1) {
+        vrna_log_error(
+          "Boltzmann sampling for multiple interacting sequences not implemented (yet)!");
+        goto exit_fail;
+      }
       print_fasta_header(output, rec_id);
 
-      fprintf(output, "%s\n", rec_sequence);
+      fprintf(output, "%s\n", orig_sequence);
 
       mfe = vrna_mfe(vc, structure);
       /* rescale Boltzmann factors according to predicted MFE */
       vrna_exp_params_rescale(vc, &mfe);
+
+      vrna_mx_mfe_free(vc);
+
       /* ignore return value, we are not interested in the free energy */
       ens_en  = vrna_pf(vc, structure);
       kT      = vc->exp_params->kT / 1000.;
@@ -495,13 +566,14 @@ main(int  argc,
     else {
       vrna_subopt_solution_t  *zr;
 
-      if (vc->cutpoint != -1)
-        vrna_message_error("Sorry, zuker subopts not yet implemented for cofold");
-
+      if (vc->strands > 1) {
+        vrna_log_error("Sorry, zuker subopts not yet implemented for cofold");
+        goto exit_fail;
+      }
       int                     i;
       print_fasta_header(output, rec_id);
 
-      fprintf(output, "%s\n", rec_sequence);
+      fprintf(output, "%s\n", orig_sequence);
 
       zr = vrna_subopt_zuker(vc);
 
@@ -542,13 +614,13 @@ main(int  argc,
 
     free(v_file_name);
 
-    if (with_shapes || (constraints_file && (!batch)))
+    if ((probing_data) || (constraints_file && (!batch)))
       break;
 
     /* print user help for the next round if we get input from tty */
     if (istty) {
       if (!zuker)
-        print_comment(stdout, "Use '&' to connect 2 sequences that shall form a complex.");
+        print_comment(stdout, "Use '&' to connect sequences that shall form a complex.");
 
       if (fold_constrained) {
         vrna_message_constraint_options(
@@ -565,14 +637,27 @@ main(int  argc,
   if (infile && input)
     fclose(input);
 
-  free(shape_method);
-  free(shape_conversion);
   free(filename_delim);
   vrna_commands_free(commands);
 
+  if (mod_params) {
+    for (vrna_sc_mod_param_t *ptr = mod_params; *ptr != NULL; ptr++)
+      vrna_sc_mod_parameters_free(*ptr);
+
+    free(mod_params);
+  }
+
   free_id_data(id_control);
 
+  probing_data_free(probing_data);
+
+  if (vrna_log_fp() != stderr)
+    fclose(vrna_log_fp());
+
   return EXIT_SUCCESS;
+
+  exit_fail:
+    return EXIT_FAILURE;
 }
 
 
